"""Solver functions."""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from scipy.linalg.lapack import get_lapack_funcs
from scipy.sparse import csr_array


def solve_linear_equation(A: NDArray, b: NDArray) -> NDArray:
    """Solve linear equation using lapack in scipy.

    numpy and scipy implementations
    x = np.linalg.solve(A, b)
    x = scipy.linalg.solve(A, b, check_finite=False, assume_a='pos')

    """
    (posv,) = get_lapack_funcs(("posv",), (A, b))
    _, x, _ = posv(A, b, lower=False, overwrite_a=False, overwrite_b=False)
    return x


def fit(X: NDArray, y: NDArray) -> NDArray:
    """Solve a normal equation in least-squares.

    (X.T @ X) @ coefs = X.T @ y

    n_samples, n_features = X.shape

    """
    coefs = solve_linear_equation(X.T @ X, X.T @ y)
    return coefs


def get_batch_slice(n_data: int, batch_size: int) -> tuple[list[int], list[int]]:
    """Calculate slice indices for a given batch size."""
    begin_batch = list(range(0, n_data, batch_size))
    if len(begin_batch) > 1:
        end_batch = list(begin_batch[1:]) + [n_data]
        if (end_batch[-1] - end_batch[-2]) < batch_size // 5:
            end_batch[-2] = end_batch[-1]
            begin_batch = begin_batch[:-1]
            end_batch = end_batch[:-1]
    else:
        end_batch = [n_data]
    return begin_batch, end_batch


def get_displacement_sparse_matrix(
    atoms: NDArray,
    displacements: NDArray,
    n_atom: int,
    tol: float = 1e-15,
) -> csr_array:
    """Return sparse matrix with displacements.

    Parameter
    ---------
    atoms: Indices of atoms displaced, shape = (n_snapshot).
    displacements: Displacement vectors, shape = (n_snapshot, 3).
    n_atom: Number of atoms in structure.
    tol: Tolerance value for defining nonzero elements.
    """
    if atoms.shape[0] != displacements.shape[0]:
        raise RuntimeError("Sizes of atoms and displacements are inconsistent.")

    N3 = n_atom * 3
    nonzero = np.abs(displacements) > tol
    rows, cols = np.where(nonzero)
    cols += atoms[rows] * 3

    mat = csr_array(
        (displacements[nonzero], (rows, cols)),
        shape=(displacements.shape[0], N3),
        dtype="double",
    )
    return mat



import numpy as np


def generate_VASP_structure(structure, scaled=False, supercell=(1, 1, 1)):

    cell = structure.get_cell(supercell=supercell)

    types = structure.get_atomic_elements(supercell=supercell)

    atom_type_unique = np.unique(types, return_index=True)

    # To use unique without sorting
    sort_index = np.argsort(atom_type_unique[1])
    elements = np.array(atom_type_unique[0])[sort_index]
    elements_count= np.diff(np.append(np.array(atom_type_unique[1])[sort_index], [len(types)]))


    vasp_POSCAR = 'Generated using dynaphopy\n'
    vasp_POSCAR += '1.0\n'
    for row in cell:
        vasp_POSCAR += '{0:20.10f} {1:20.10f} {2:20.10f}\n'.format(*row)
    vasp_POSCAR += ' '.join(elements)
    vasp_POSCAR += ' \n'
    vasp_POSCAR += ' '.join([str(i) for i in elements_count])

    if scaled:
        scaled_positions = structure.get_scaled_positions(supercell=supercell)
        vasp_POSCAR += '\nDirect\n'
        for row in scaled_positions:
            vasp_POSCAR += '{0:15.15f}   {1:15.15f}   {2:15.15f}\n'.format(*row)

    else:
        positions = structure.get_positions(supercell=supercell)
        vasp_POSCAR += '\nCartesian\n'
        for row in positions:
            vasp_POSCAR += '{0:20.10f} {1:20.10f} {2:20.10f}\n'.format(*row)

    return vasp_POSCAR


def generate_LAMMPS_structure(structure, supercell=(1, 1, 1)):

    types = structure.get_atomic_elements(supercell=supercell)

    type_index_unique = np.unique(types, return_index=True)[1]

    # To use unique without sorting
    sort_index = np.argsort(type_index_unique)
    type_index_unique = np.array(type_index_unique)[sort_index]

    count_index_unique = np.diff(np.append(type_index_unique, [len(types)]))

    atom_index = []
    for i, index in enumerate(count_index_unique):
        atom_index += [i for j in range(index)]

    atom_index_unique = np.unique(atom_index, return_index=True)[1]

    index_dict = {}
    for i, v in enumerate(atom_index_unique):
        index_dict[types[v]] = i + 1

    masses = structure.get_masses(supercell=supercell)
    charges = structure.get_charges(supercell=supercell)

    positions = structure.get_positions(supercell=supercell)
    number_of_atoms = len(positions)

    lammps_data_file = 'Generated using dynaphopy\n\n'
    lammps_data_file += '{0} atoms\n\n'.format(number_of_atoms)
    lammps_data_file += '{0} atom types\n\n'.format(len(atom_index_unique))

    cell = structure.get_cell(supercell=supercell)

    # generate lammps oriented lattice vectors
    ax = np.linalg.norm(cell[0])
    bx = np.dot(cell[1], cell[0]/np.linalg.norm(cell[0]))
    by = np.linalg.norm(np.cross(cell[0]/np.linalg.norm(cell[0]), cell[1]))
    cx = np.dot(cell[2], cell[0]/np.linalg.norm(cell[0]))
    cy = (np.dot(cell[1], cell[2])- bx*cx)/by
    cz = np.sqrt(np.dot(cell[2],cell[2])-pow(cx,2)-pow(cy,2))

    # get rotation matrix (poscar->lammps) from lattice vectors matrix
    lammps_cell = np.array([[ax, bx, cx],[0, by, cy],[0,0,cz]])
    trans_cell = np.dot(np.linalg.inv(cell), lammps_cell.T)

    # rotate positions to lammps orientation
    positions = np.dot(positions, trans_cell)

    # build lammps lattice vectors
    a, b, c, alpha, beta, gamma = structure.get_cell_parameters(supercell=supercell)

    xhi = a
    xy = b * np.cos(gamma)
    xz = c * np.cos(beta)
    yhi = np.sqrt(pow(b,2)-pow(xy,2))
    yz = (b*c*np.cos(alpha)-xy * xz)/yhi
    zhi = np.sqrt(pow(c,2)-pow(xz,2)-pow(yz,2))

    xhi = xhi + max(0,0, xy, xz, xy+xz)
    yhi = yhi + max(0,0, yz)

    if xy > 0:
        xhi -= xy
    if xz > 0:
        xhi -= xz
    if yz > 0:
        yhi -= yz

    # write lammpstrj file
    lammps_data_file += '\n{0:20.10f} {1:20.10f} xlo xhi\n'.format(0, xhi)
    lammps_data_file += '{0:20.10f} {1:20.10f} ylo yhi\n'.format(0, yhi)
    lammps_data_file += '{0:20.10f} {1:20.10f} zlo zhi\n'.format(0, zhi)
    lammps_data_file += '{0:20.10f} {1:20.10f} {2:20.10f} xy xz yz\n\n'.format(xy, xz, yz)

    lammps_data_file += 'Masses\n\n'

    for i, index in enumerate(atom_index_unique):
        lammps_data_file += '{0} {1:20.10f} \n'.format(i+1, masses[index])

    lammps_data_file += '\nAtoms\n\n'

    if charges is not None:
        for i, row in enumerate(positions):
            lammps_data_file += '{0} {1} {2} {3:20.10f} {4:20.10f} {5:20.10f}\n'.format(i + 1, index_dict[types[i]],
                                                                                        charges[i], row[0], row[1], row[2])
    else:
        for i, row in enumerate(positions):
            lammps_data_file += '{0:10} {1} {2:20.10f} {3:20.10f} {4:20.10f}\n'.format(i+1, index_dict[types[i]], row[0],row[1],row[2])
            #lammps_data_file += '{0} {1:20.10f} {2:20.10f} {3:20.10f}\n'.format(index_dict[types[i]], row[0],row[1],row[2])

    return lammps_data_file

if __name__ == "__main__":

    import macer.externals.dynaphopy_bundled.interface.iofile as reading
    input_parameters = reading.read_parameters_from_input_file('/home/abel/VASP/Ag2Cu2O4/MD/input_dynaphopy')
    structure = reading.read_from_file_structure_poscar(input_parameters['structure_file_name_poscar'])
 #   print(generate_VASP_structure(structure))
    print(generate_LAMMPS_structure(structure))


# pylint: disable=cyclic-import, wrong-import-position
# ruff: noqa: E402
"""
The seekpath module contains routines to get automatically the
path in a 3D Brillouin zone to plot band structures.

Author: Giovanni Pizzi, PSI (2016-onwards)
Maintainer: Timo Reents, PSI (2025-onwards)

Licence: MIT License, see LICENSE.txt file
"""

__version__ = '2.2.0'
__author__ = 'Giovanni Pizzi, PSI'
__maintainer__ = 'Timo Reents, PSI'
__copyright__ = (
    'Copyright (c), 2016-2026, Giovanni Pizzi, PAUL SCHERRER INSTITUT '
    '(Laboratory for Materials Simulations), EPFL '
    '(Theory and Simulation of Materials (THEOS) and National Centre '
    'for Computational Design and Discovery of Novel Materials '
    '(NCCR MARVEL)), Switzerland.'
)
__credits__ = ['Yoyo Hinuma', 'Jae-Mo Lihm']
__license__ = 'MIT license'
__paper__ = (
    'Y. Hinuma, G. Pizzi, Y. Kumagai, F. Oba, I. Tanaka, '
    'Band structure diagram paths based on crystallography, '
    'Comp. Mat. Sci. 128, 140 (2017). DOI: 10.1016/j.commatsci.2016.10.015'
)


class SupercellWarning(UserWarning):
    """
    A warning issued when the cell is an undistorted supercell of a smaller
    unit cell, and the kpoint path for a non-standardized cell (i.e., for the
    original cell) is requested.
    """


from .getpaths import (
    get_path,
    get_explicit_k_path,
    get_path_orig_cell,
    get_explicit_k_path_orig_cell,
)

from .hpkot import EdgeCaseWarning, SymmetryDetectionError
from .brillouinzone import brillouinzone


__all__ = (
    'get_path',
    'get_explicit_k_path',
    'get_path_orig_cell',
    'get_explicit_k_path_orig_cell',
    'EdgeCaseWarning',
    'SymmetryDetectionError',
    'SupercellWarning',
    'brillouinzone',
)



import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as stats

kb_boltzmann = 0.831446  # u * A^2 / ( ps^2 * K )


def boltzmann_distribution(trajectory, parameters):

    print("\nMaxwell-Boltzmann distribution analysis")
    print("----------------------------------------------")

    velocity = np.reshape(np.linalg.norm(trajectory.get_velocity_mass_average(), axis=2), -1)

    average = np.average(velocity)
    deviation = np.std(velocity)
    print('Average: {0:3.7e} Angstrom/ps'.format(average))
    print('Deviation: {0:3.7e} Angstrom/ps'.format(deviation))
    maxwell = stats.maxwell

    params = maxwell.fit(velocity, floc=0, scale=np.average([average, deviation]))
    print('Distribution parameter: {0:3.7e} Angstrom/ps'.format(params[1]))

    temperature = pow(params[1], 2)/ kb_boltzmann
    print('Temperature fit: {0:7.6f} K'.format(temperature))

    if not parameters.silent:
        x = np.linspace(0, float(average + 3 * deviation), 100)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        plt.xlabel('Velocity [Angstrom/ps]')
        ax.plot(x, maxwell.pdf(x, *params), lw=3)
        ax.text(0.95, 0.90, 'Temperature: {0:7.1f} K'.format(temperature),
                verticalalignment='bottom',
                horizontalalignment='right',
                transform=ax.transAxes,
                fontsize=15)

        fig.suptitle('Velocity distribution')
        ax.hist(velocity, bins=parameters.number_of_bins_histogram, density=True)

        plt.show()



    return temperature


"""SpaceGroup operations."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import warnings
from typing import Any, cast

import numpy as np
from numpy.typing import ArrayLike, NDArray

import macer.externals.spglib_bundled as spglib

try:
    from . import _spglib
except ImportError:
    try:
        from spglib import _spglib
    except ImportError:
        import _spglib

from ._compat.typing import TypeAlias
from ._compat.warnings import deprecated
from .error import _set_no_error, _set_or_throw_error
from .utils import Cell, DictInterface, Lattice, Numbers, Positions, _expand_cell

__all__ = [
    "SpgCell",
    "SpglibDataset",
    "SpaceGroupType",
    "get_symmetry",
    "get_symmetry_dataset",
    "get_spacegroup",
    "get_spacegroup_type",
    "get_spacegroup_type_from_symmetry",
    "get_symmetry_from_database",
    "get_hall_number_from_symmetry",
]


SpgCell: TypeAlias = tuple[Lattice, Positions, Numbers]
"""
Crystal structure represented by a tuple of (:py:type:`spglib.utils.Lattice`,
:py:type:`spglib.utils.Positions`, :py:type:`spglib.utils.Numbers`).
"""


@dataclasses.dataclass(eq=False, frozen=True)
class SpglibDataset(DictInterface):
    """Spglib dataset information.

    .. versionadded:: 1.9.4
        The member 'choice' is added.
    .. versionadded:: 2.5.0
    """

    number: int
    """International space group number"""
    hall_number: int
    """Hall number.

    This number is used in
        :func:`get_symmetry_from_database` and
        :func:`get_spacegroup_type`.
    """
    international: str
    """International symbol"""
    hall: str
    """Hall symbol"""
    choice: str
    """Centring, origin, basis vector setting"""
    transformation_matrix: NDArray[np.double]
    """Transformation matrix from input lattice to standardized lattice.

    shape=(3, 3), order='C', dtype='double'

    .. code-block::

        L^original = L^standardized * Tmat.

    See the detail at :ref:`dataset_origin_shift_and_transformation`.
    """
    origin_shift: NDArray[np.double]
    """Origin shift from standardized to input origin.

    shape=(3,), dtype='double'

    See the detail at :ref:`dataset_origin_shift_and_transformation`.
    """
    rotations: NDArray[np.intc]
    """Matrix (rotation) parts of space group operations.

    shape=(n_operations, 3, 3), order='C', dtype='intc'

    .. note::

        Space group operations are obtained by

        .. code-block:: python

            [(r,t) for r, t in zip(rotations, translations)]

    See also :func:`get_symmetry`.
    """
    translations: NDArray[np.double]
    """Vector (translation) parts of space group operations.

    shape=(n_operations, 3), order='C', dtype='double'

    .. note::

        Space group operations are obtained by

        .. code-block:: python

            [(r,t) for r, t in zip(rotations, translations)]

    See also :func:`get_symmetry`.
    """
    wyckoffs: list[str]
    """Wyckoff letters corresponding to the space group type."""
    site_symmetry_symbols: list[str]
    """Site symmetry symbols corresponding to the space group type."""
    crystallographic_orbits: NDArray[np.intc]
    """Symmetrically equivalent atoms, where 'symmetrically' means the space group
    operations corresponding to the space group type.

    This is very similar to ``equivalent_atoms``. See the difference explained in
    ``equivalent_atoms``

    shape=(n_atoms,), dtype='intc'
    """
    equivalent_atoms: NDArray[np.intc]
    """Symmetrically equivalent atoms, where 'symmetrically' means found symmetry
    operations.

    shape=(n_atoms,), dtype='intc'

    In spglib, symmetry operations are given for the input cell.
    When a non-primitive cell is inputted and the lattice made by the non-primitive
    basis vectors breaks its point group, the found symmetry operations may not
    correspond to the crystallographic space group type.
    """
    # Primitive cell
    primitive_lattice: NDArray[np.double]
    """Basis vectors a, b, c of a primitive cell in row vectors.

    shape=(3, 3), order='C', dtype='double'

    This is the primitive cell found inside spglib before applying standardization.
    Therefore, the values are distorted with respect to found space group type.
    """
    mapping_to_primitive: NDArray[np.intc]
    """Atom index mapping from original cell to the primitive cell of
    ``primitive_lattice``.

    shape=(n_atoms), dtype='intc'
    """
    # Idealized standardized unit cell
    std_lattice: NDArray[np.double]
    """Basis vectors a, b, c of an idealized standardized cell in row vectors.

    shape=(3, 3), order='C', dtype='double'
    """
    std_positions: NDArray[np.double]
    """Positions of atoms in the standardized cell in fractional coordinates.

    shape=(n_atoms, 3), order='C', dtype='double'
    """
    std_types: NDArray[np.intc]
    """Identity numbers of atoms in the standardized cell.

    shape=(n_atoms,), dtype='intc'
    """
    std_rotation_matrix: NDArray[np.double]
    """Rigid rotation matrix to rotate from standardized basis vectors to idealized
    standardized orthonormal basis vectors.

    shape=(3, 3), order='C', dtype='double'

    .. code-block::

        L^idealized = R * L^standardized.

    See the detail at :ref:`dataset_idealized_cell`.
    """
    std_mapping_to_primitive: NDArray[np.intc]
    """``std_positions`` index mapping to those of atoms of the primitive cell in the
    standardized cell."""
    pointgroup: str
    """Pointgroup symbol in Hermann-Mauguin notation."""


@dataclasses.dataclass(eq=True, frozen=True)
class SpaceGroupType(DictInterface):
    """Space group type information.

    More details are found at :ref:`spglib-dataset`.

    .. versionchanged:: 2.0
        ``hall_number`` member is added.
    .. versionadded:: 2.5.0
    """

    number: int
    """International space group number"""
    international_short: str
    """International short symbol"""
    international_full: str
    """International full symbol"""
    international: str
    """International symbol"""
    schoenflies: str
    """Schoenflies symbol"""
    hall_number: int
    """Hall symbol serial number"""
    hall_symbol: str
    """Hall symbol"""
    choice: str
    """Centring, origin, basis vector setting"""
    pointgroup_international: str
    """International symbol of crystallographic point group"""
    pointgroup_schoenflies: str
    """Schoenflies symbol of crystallographic point group"""
    arithmetic_crystal_class_number: int
    """Arithmetic crystal class number"""
    arithmetic_crystal_class_symbol: str
    """Arithmetic crystal class symbol"""


def get_symmetry(
    cell: Cell,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
    mag_symprec: float = -1.0,
    is_magnetic: bool = True,
) -> dict[str, Any] | None:
    r"""Find symmetry operations from a crystal structure and site tensors.

    .. warning::
        :func:`get_symmetry` with ``is_magnetic=True`` is deprecated at version 2.0.

    Use :func:`get_magnetic_symmetry` for magnetic symmetry search.

    Parameters
    ----------
    cell : tuple
        Crystal structure given in tuple.
        It has to follow the following form,
        (basis vectors, atomic points, types in integer numbers, ...)

        - basis vectors : array_like
            shape=(3, 3), order='C', dtype='double'

            .. code-block:: python

                [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

        - atomic points : array_like
            shape=(num_atom, 3), order='C', dtype='double'

            Atomic position vectors with respect to basis vectors, i.e.,
            given in  fractional coordinates.
        - types : array_like
            shape=(num_atom, ), dtype='intc'

            Integer numbers to distinguish species.
        - optional data :
            case-I: Scalar
                shape=(num_atom, ), dtype='double'

                Each atomic site has a scalar value. With is_magnetic=True,
                values are included in the symmetry search in a way of
                collinear magnetic moments.
            case-II: Vectors
                shape=(num_atom, 3), order='C', dtype='double'

                Each atomic site has a vector. With is_magnetic=True,
                vectors are included in the symmetry search in a way of
                non-collinear magnetic moments.
    symprec : float
        Symmetry search tolerance in the unit of length.
    angle_tolerance : float
        Symmetry search tolerance in the unit of angle deg.
        Normally it is not recommended to use this argument.
        See a bit more detail at :ref:`variables_angle_tolerance`.
        If the value is negative, an internally optimized routine is used to
        judge symmetry.
    mag_symprec : float
        Tolerance for magnetic symmetry search in the unit of magnetic moments.
        If not specified, use the same value as symprec.
    is_magnetic : bool
        When optional data (4th element of cell tuple) is given in case-II,
        the symmetry search is performed considering magnetic symmetry, which
        may be corresponding to that for non-collinear calculation. Default is
        True, but this does nothing unless optional data is supplied.

    Returns
    -------
    symmetry: dict
        Rotation parts and translation parts of symmetry operations are represented
        with respect to basis vectors.
        When the search failed, :code:`None` is returned.

        - 'rotations' : ndarray
            shape=(num_operations, 3, 3), order='C', dtype='intc'

            Rotation (matrix) parts of symmetry operations
        - 'translations' : ndarray
            shape=(num_operations, 3), dtype='double'

            Translation (vector) parts of symmetry operations
        - 'time_reversals': ndarray (exists when the optional data is given)
            shape=(num_operations, ), dtype='bool\_'

            Time reversal part of magnetic symmetry operations.
            True indicates time reversal operation, and False indicates
            an ordinary operation.
        - 'equivalent_atoms' : ndarray
            shape=(num_atoms, ), dtype='intc'

            A mapping table of atoms to symmetrically independent atoms.
            This is used to find symmetrically equivalent atoms.
            The numbers contained are the indices of atoms starting from 0,
            i.e., the first atom is numbered as 0, and
            then 1, 2, 3, ... :code:`np.unique(equivalent_atoms)` gives representative
            symmetrically independent atoms. A list of atoms that are
            symmetrically equivalent to some independent atom (here for example 1
            is in :code:`equivalent_atom`) is found by
            :code:`np.where(equivalent_atom=1)[0]`.

    Notes
    -----
    The orders of the rotation matrices and the translation
    vectors correspond with each other, e.g. , the second symmetry
    operation is organized by the set of the second rotation matrix and second
    translation vector in the respective arrays. Therefore a set of
    symmetry operations may obtained by

    .. code-block:: python

        [(r, t) for r, t in zip(dataset['rotations'], dataset['translations'])]

    The operations are given with respect to the fractional coordinates
    (not for Cartesian coordinates). The rotation matrix and translation
    vector are used as follows:

    .. code-block::

        new_vector[3x1] = rotation[3x3] * vector[3x1] + translation[3x1]

    The three values in the vector are given for the a, b, and c axes,
    respectively.

    """
    _set_no_error()

    _, _, _, magmoms = _expand_cell(cell)

    if magmoms is None:
        # Get symmetry operations without on-site tensors (i.e. normal crystal)
        try:
            dataset = get_symmetry_dataset(
                cast(SpgCell, cell),
                symprec=symprec,
                angle_tolerance=angle_tolerance,
                _throw=True,
            )
        except Exception as exc:
            _set_or_throw_error(exc)
            return None
        return {
            "rotations": dataset.rotations,
            "translations": dataset.translations,
            "equivalent_atoms": dataset.equivalent_atoms,
        }
    else:
        warnings.warn(
            "Use get_magnetic_symmetry() for cell with magnetic moments.",
            DeprecationWarning,
            stacklevel=1,
        )
        try:
            return spglib.get_magnetic_symmetry(
                cast(spglib.MsgCell, cell),
                symprec=symprec,
                angle_tolerance=angle_tolerance,
                mag_symprec=mag_symprec,
                is_axial=None,
                with_time_reversal=is_magnetic,
                _throw=True,
            )
        except Exception as exc:
            _set_or_throw_error(exc)
            return None


def get_symmetry_dataset(
    cell: SpgCell,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
    hall_number: int = 0,
    _throw: bool = False,
) -> SpglibDataset | None:
    """Search symmetry dataset from an input cell.

    Parameters
    ----------
    cell, symprec, angle_tolerance:
        See :func:`get_symmetry`.
    hall_number : int
        If a serial number of Hall symbol (>0) is given, the database
        corresponding to the Hall symbol is made.

        The mapping from Hall symbols to a space-group-type is the many-to-one mapping.
        Without specifying this
        option (i.e., in the case of ``hall_number=0``), always the first one
        (the smallest serial number corresponding to the space-group-type in
        [list of space groups (Seto's web site)](https://yseto.net/en/sg/sg1))
        among possible choices and settings is chosen as default. This
        argument is useful when the other choice (or setting) is expected to
        be hooked.

        This affects to the obtained values of `international`,
        `hall`, `choice`, `transformation_matrix`,
        `origin shift`, `wyckoffs`, `std_lattice`, `std_positions`,
        `std_types` and `std_rotation_matrix`, but not to `rotations`
        and `translations` since the later set is defined with respect to
        the basis vectors of user's input (the `cell` argument).

        See also :ref:`dataset_spg_get_dataset_spacegroup_type`.

    Returns
    -------
    dataset: :class:`SpglibDataset` | None
        If it fails, None is returned. Otherwise a dictionary is returned.
        More details are found at :ref:`spglib-dataset`.

    """
    _set_no_error(_throw)

    lattice, positions, numbers, _ = _expand_cell(cell)

    try:
        spg_ds = _spglib.dataset(
            lattice,
            positions,
            numbers,
            int(hall_number),
            float(symprec),
            float(angle_tolerance),
        )
    except Exception as exc:
        _set_or_throw_error(exc, _throw)
        return None

    return SpglibDataset(**spg_ds)


def get_spacegroup(
    cell: SpgCell,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
    symbol_type: int = 0,
) -> str | None:
    """Return space group in international table symbol and number as a string.

    With ``symbol_type=1``, Schoenflies symbol is given instead of international symbol.

    :rtype: str | None
    :return:
        If it fails, None is returned.
    """
    _set_no_error()

    try:
        dataset = get_symmetry_dataset(
            cell,
            symprec=symprec,
            angle_tolerance=angle_tolerance,
            _throw=True,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None

    try:
        spg_type = get_spacegroup_type(dataset.hall_number, _throw=True)
    except Exception as exc:
        _set_or_throw_error(exc)
        return None

    if symbol_type == 1:
        return f"{spg_type.schoenflies} ({dataset.number})"
    else:
        return f"{spg_type.international_short} ({dataset.number})"


def get_spacegroup_type(
    hall_number: int,
    _throw: bool = False,
) -> SpaceGroupType | None:
    """Translate Hall number to space group type information. If it fails, return None.

    This function allows to directly access to the space-group-type database
    in spglib (spg_database.c).
    To specify the space group type with a specific choice, ``hall_number`` is used.
    The definition of ``hall_number`` is found at
    :ref:`dataset_spg_get_dataset_spacegroup_type`.

    :param hall_number: Serial number for Hall symbol
    :return: :class:`SpaceGroupType` or None

    .. versionadded:: 1.9.4
    """
    _set_no_error(_throw)

    try:
        spg_type = _spglib.spacegroup_type(int(hall_number))
    except Exception as exc:
        _set_or_throw_error(exc, _throw)
        return None
    return SpaceGroupType(**spg_type)


def get_spacegroup_type_from_symmetry(
    rotations: ArrayLike[np.intc],
    translations: ArrayLike[np.double],
    lattice: ArrayLike[np.double] | None = None,
    symprec: float = 1e-5,
) -> SpaceGroupType | None:
    """Return space-group type information from symmetry operations.

    This is expected to work well for the set of symmetry operations whose
    distortion is small. The aim of making this feature is to find
    space-group-type for the set of symmetry operations given by the other
    source than spglib.

    Parameters
    ----------
    rotations : array_like
        Matrix parts of space group operations.
        shape=(n_operations, 3, 3), order='C', dtype='intc'
    translations : array_like
        Vector parts of space group operations.
        shape=(n_operations, 3), order='C', dtype='double'
    lattice : array_like, optional
        Basis vectors a, b, c given in row vectors. Default is None, which gives
        unit matrix. This should be the same as that used to find ``rotations``
        and ``translations``. If it is unknown, the cubic basis vector may be a
        possible choice unless the ``rotations`` and ``translations`` were
        obtained for an unusual (very oblique) choice of basis vectors.
        shape=(3, 3), order='C', dtype='double'
    symprec: float
        See :func:`get_symmetry`.

    Returns
    -------
    spacegroup_type : :class:`SpaceGroupType` or None

    Notes
    -----
    .. versionadded:: 2.0

    This is the replacement of :func:`get_hall_number_from_symmetry`.

    """
    r = np.array(rotations, dtype="intc", order="C")
    t = np.array(translations, dtype="double", order="C")
    if lattice is None:
        _lattice = np.eye(3, dtype="double", order="C")
    else:
        _lattice = np.array(np.transpose(lattice), dtype="double", order="C")

    _set_no_error()

    try:
        spg_type = _spglib.spacegroup_type_from_symmetry(r, t, _lattice, float(symprec))
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return SpaceGroupType(**spg_type)


def get_symmetry_from_database(hall_number: int) -> dict[str, Any] | None:
    """Return symmetry operations corresponding to a Hall symbol. If fails, return None.

    Parameters
    ----------
    hall_number : int
        The Hall symbol is given by the serial number in between 1 and 530.
        The definition of ``hall_number`` is found at
        :ref:`dataset_spg_get_dataset_spacegroup_type`.

    Returns
    -------
    symmetry : dict
        - rotations
            Rotation parts of symmetry operations corresponding to ``hall_number``.
        - translations
            Translation parts of symmetry operations corresponding to ``hall_number``.

    """
    _set_no_error()

    rotations = np.zeros((192, 3, 3), dtype="intc")
    translations = np.zeros((192, 3), dtype="double")
    try:
        num_sym = _spglib.symmetry_from_database(
            rotations,
            translations,
            int(hall_number),
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return {
        "rotations": np.array(rotations[:num_sym], dtype="intc", order="C"),
        "translations": np.array(translations[:num_sym], dtype="double", order="C"),
    }


@deprecated("Use get_spacegroup_type_from_symmetry instead")
def get_hall_number_from_symmetry(
    rotations: ArrayLike[np.intc],
    translations: ArrayLike[np.double],
    symprec: float = 1e-5,
) -> int | None:
    """Hall number is obtained from a set of symmetry operations. If fails, return None.

    .. deprecated:: 2.0
        Replaced by {func}`get_spacegroup_type_from_symmetry`.

    Return one of ``hall_number`` corresponding to a space-group type of the given
    set of symmetry operations. When multiple ``hall_number`` exist for the
    space-group type, the smallest one (the first description of the space-group
    type in International Tables for Crystallography) is chosen. The definition of
    ``hall_number`` is found at :ref:`dataset_spg_get_dataset_spacegroup_type` and
    the corresponding space-group-type information is obtained through
    {func}`get_spacegroup_type`.

    This is expected to work well for the set of symmetry operations whose
    distortion is small. The aim of making this feature is to find
    space-group-type for the set of symmetry operations given by the other
    source than spglib.

    Note that the definition of ``symprec`` is
    different from usual one, but is given in the fractional
    coordinates and so it should be small like ``1e-5``.
    """
    _set_no_error()

    r = np.array(rotations, dtype="intc", order="C")
    t = np.array(translations, dtype="double", order="C")
    try:
        hall_number = _spglib.hall_number_from_symmetry(r, t, float(symprec))
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return hall_number



"""
Macer: Machine-learning accelerated Atomic Computational Environment for automated Research workflows
Copyright (c) 2025 The Macer Package Authors
Author: Soungmin Bae <soungminbae@gmail.com>
License: MIT
"""

import sys
import os
import shutil
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
import yaml
from argparse import Namespace
from matplotlib import pyplot as plt
import numpy as np
from numpy.linalg import det
from monty.serialization import loadfn
from pymatgen.core import Structure, Composition, IStructure
from pymatgen.analysis.phase_diagram import PhaseDiagram, PDEntry

from pydefect.defaults import defaults
from pydefect.analyzer.make_defect_structure_info import MakeDefectStructureInfo
from pydefect.cli.make_defect_vesta_file import MakeDefectVestaFile
from pydefect.defaults import defaults as pydefect_defaults
from pydefect.input_maker.manual_supercell_maker import ManualSupercellMaker
from pydefect.input_maker.supercell import Supercell
from pydefect.input_maker.supercell_info import SimpleSite, SupercellInfo
from vise.util.structure_symmetrizer import StructureSymmetrizer, Site

from macer.calculator.factory import get_available_ffs, get_calculator
from macer.relaxation.optimizer import relax_structure
from macer.utils.logger import Logger
from macer.io.writers import write_pydefect_dummy_files
from macer.defaults import DEFAULT_MODELS, _macer_root, _model_root, DEFAULT_DEVICE, resolve_model_path

# Monkeypatch for pydefect 0.8.x which crashes if sites_yaml_filename is None
try:
    from pydefect.input_maker import manual_supercell_maker
    from pydefect.cli import main_functions
    
    _original_make_sites = manual_supercell_maker.make_sites_from_yaml_file

    def _patched_make_sites_from_yaml_file(filename):
        if filename is None:
            return {}
        return _original_make_sites(filename)

    manual_supercell_maker.make_sites_from_yaml_file = _patched_make_sites_from_yaml_file
    main_functions.make_sites_from_yaml_file = _patched_make_sites_from_yaml_file
except ImportError:
    pass

def generate_composition_energies_direct(directories, yaml_filename="composition_energies.yaml"):
    """
    Directly generates composition_energies.yaml from calc_results.json files 
    in the specified directories, bypassing vise/OUTCAR parsing.
    """
    data = {}
    for d in directories:
        d_path = Path(d)
        calc_results_path = d_path / "calc_results.json"
        
        if not calc_results_path.exists():
            print(f"Warning: {calc_results_path} not found. Skipping {d_path.name}.")
            continue
        
        try:
            # Load calc_results.json generated by macer
            calc_results = loadfn(calc_results_path)
            
            # Check if calc_results is a dict or an object
            if isinstance(calc_results, dict):
                energy = calc_results.get("energy")
                structure_dict = calc_results.get("structure")
                if structure_dict:
                    structure = Structure.from_dict(structure_dict)
                else:
                    structure = None
            else:
                # Assume it's a CalcResults object
                energy = getattr(calc_results, "energy", None)
                structure = getattr(calc_results, "structure", None)
            
            if energy is None or structure is None:
                print(f"Warning: Incomplete data in {calc_results_path}. Skipping.")
                continue
                
            # Use formula with spaces removed (e.g. "Zn2 O2" -> "Zn2O2") as key
            formula = structure.composition.formula.replace(" ", "")
            
            # If multiple phases have the same formula, we should theoretically keep the most stable one.
            # However, pydefect usually handles competing phases by ensuring they are distinct or 
            # by processing them such that the most stable ones define the hull.
            # Here we simply store it. If duplicates exist, the last one visited overwrites.
            # Ideally, directories should correspond to unique stable phases.
            
            data[formula] = {
                "energy": float(energy),
                "source": str(d_path.absolute())
            }
        except Exception as e:
            print(f"Error processing {d_path.name}: {e}")
            
    if data:
        with open(yaml_filename, "w") as f:
            yaml.dump(data, f)
        print(f"Directly generated {yaml_filename} from {len(data)} entries.")
        return True
    else:
        print(f"No data collected for {yaml_filename}.")
        return False

def get_unique_dir_name(base_name):
    if not Path(base_name).exists():
        return Path(base_name)
    i = 1
    while True:
        new_name = f"{base_name}-NEW{i:02d}"
        if not Path(new_name).exists():
            return Path(new_name)
        i += 1

def _as_float_list(seq):
    return [float(x) for x in seq]


def dim_to_matrix(dim):
    if dim is None:
        return None
    d = list(dim)
    if len(d) == 3:
        return [d[0], d[1], d[2]]
    if len(d) == 9:
        return d
    raise ValueError(
        "--dim must be 3 (diag) or 9 (full matrix) integers.")


def _dim_to_matrix_3x3(dim):
    if dim is None:
        return None
    d = list(dim)
    if len(d) == 3:
        return [[d[0], 0, 0], [0, d[1], 0], [0, 0, d[2]]]
    if len(d) == 9:
        return [d[0:3], d[3:6], d[6:9]]
    raise ValueError(
        "--dim must be 3 (diag) or 9 (full matrix) integers.")


def _make_simple_sites_from_structure(structure):
    counts = {}
    sites = {}
    for idx, site in enumerate(structure):
        elem = site.specie.symbol
        counts[elem] = counts.get(elem, 0) + 1
        name = f"{elem}{counts[elem]}"
        sites[name] = SimpleSite(elem, idx)
    return sites


def _make_irreducible_sites(structure, matrix,
                            symprec=pydefect_defaults.symmetry_length_tolerance,
                            angle_tolerance=pydefect_defaults.symmetry_angle_tolerance):
    symmetrizer = StructureSymmetrizer(structure,
                                       symprec=symprec,
                                       angle_tolerance=angle_tolerance)
    multiplicity = int(round(det(matrix)))

    sites = {}
    for site_name, site in symmetrizer.sites.items():
        atoms = []
        for i in site.equivalent_atoms:
            atoms.extend(list(range(i * multiplicity, (i + 1) * multiplicity)))
        sites[site_name] = Site(element=site.element,
                                wyckoff_letter=site.wyckoff_letter,
                                site_symmetry=site.site_symmetry,
                                equivalent_atoms=atoms)
    return symmetrizer.spglib_sym_data.international, sites


def make_manual_supercell_info(input_poscar: Path, dim, output_dir: Path = Path("."),
                               analyze_symmetry: bool = True):
    if dim is None:
        raise ValueError("--dim is required.")
    matrix = _dim_to_matrix_3x3(dim)
    input_poscar = Path(input_poscar).expanduser().resolve()
    if not input_poscar.exists():
        raise FileNotFoundError(f"{input_poscar} not found.")

    output_dir = Path(output_dir).expanduser().resolve()
    output_dir.mkdir(parents=True, exist_ok=True)

    cwd = Path.cwd()
    try:
        os.chdir(output_dir)
        # keep a local copy for traceability
        try:
            shutil.copy(str(input_poscar), "POSCAR-unitcell")
        except Exception:
            pass

        unitcell = IStructure.from_file(str(input_poscar))
        if analyze_symmetry:
            space_group, sites = _make_irreducible_sites(unitcell, matrix)
            supercell = Supercell(unitcell, matrix)
            supercell.structure.to(filename="SPOSCAR")
            supercell_info = SupercellInfo(
                structure=supercell.structure,
                space_group=space_group,
                transformation_matrix=matrix,
                sites=sites,
                unitcell_structure=unitcell
            )
            supercell_info.to_json_file()
        else:
            sites = _make_simple_sites_from_structure(unitcell)
            maker = ManualSupercellMaker(unitcell, sites=sites, matrix=matrix)
            maker.supercell.structure.to(filename="SPOSCAR")
            maker.supercell_info.to_json_file()
        print(f"Created supercell_info.json in {output_dir}")
    finally:
        os.chdir(cwd)


def plot_cpd_outputs(cpd,
                     plot_elements=None,
                     output_prefix="cpd",
                     make_pdf=True,
                     make_html=True,
                     project_elements=None):
    from pydefect.chem_pot_diag.chem_pot_diag import change_element_sequence, TargetVertices
    from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiag2DMplPlotter, ChemPotDiag3DMplPlotter
    import plotly.graph_objects as go
    def _sort_coords_2d(coords_2d):
        arr = np.array(coords_2d, dtype=float)
        if len(arr) == 0:
            return []
        center = arr.mean(axis=0)
        angles = np.arctan2(arr[:, 1] - center[1], arr[:, 0] - center[0])
        order = np.argsort(angles)
        return arr[order].tolist()

    # Align element order (only when plotting full CPD for 2D/3D)
    if plot_elements and not project_elements:
        try:
            cpd = change_element_sequence(cpd, element_sequence=plot_elements)
        except Exception as e:
            print(f"Invalid plot element order {plot_elements}: {e}")
            return False
    elif not plot_elements and not project_elements:
        cpd = change_element_sequence(cpd)

    # Ensure target vertices are available for info panel
    if cpd.target and not cpd.target_vertices_dict:
        try:
            target_vertices = cpd.to_target_vertices
            cpd.target_vertices_dict = target_vertices.vertices
        except Exception:
            pass
    if Path("target_vertices.yaml").exists():
        try:
            target_vertices = TargetVertices.from_yaml("target_vertices.yaml")
            cpd.target_vertices_dict = target_vertices.vertices
        except Exception:
            pass

    project_indices = None
    project_elements_used = None
    if project_elements:
        if len(project_elements) != 2:
            print("project_elements must have exactly 2 elements.")
            return False
        missing = [e for e in project_elements if e not in cpd.vertex_elements]
        if missing:
            print(f"Invalid project_elements {missing}. Available: {cpd.vertex_elements}")
            return False
        project_indices = [cpd.vertex_elements.index(e) for e in project_elements]
        project_elements_used = list(project_elements)

    if cpd.dim not in (2, 3) and not project_indices:
        print(f"CPD dimension {cpd.dim} is not supported for plotting. Skipping outputs.")
        print("Tip: use 2D projection for higher-dimensional systems, e.g.:")
        print("  macer util pydefect cpd -r relative_energies.yaml -t <TARGET> --projection")
        print("  macer util pydefect cpd -r relative_energies.yaml -t <TARGET> --projection-elements E1 E2")
        return False

    if make_pdf:
        if cpd.dim == 2:
            plotter = ChemPotDiag2DMplPlotter(cpd)
        elif cpd.dim == 3:
            plotter = ChemPotDiag3DMplPlotter(cpd)
        else:
            plotter = None
        if plotter is None:
            return False
        plt_obj = plotter.draw_diagram()
        plt_obj.savefig(f"{output_prefix}.pdf")
        plt.close("all")
        print(f"Generated {output_prefix}.pdf")

    if make_html:
        title = f"Chemical potential diagram of {cpd.chemical_system}"
        traces = []
        all_vertex_elements = list(cpd.vertex_elements) if cpd.vertex_elements else []
        vertex_elements = project_elements_used if project_elements_used else all_vertex_elements

        def _fmt_mu(el, chem_pot):
            val = chem_pot.get(el, None)
            return f"μ_{el}={val:.4f}" if val is not None else f"μ_{el}=N/A"

        vertex_entries = []
        if cpd.target_vertices_dict:
            for label in sorted(cpd.target_vertices_dict.keys()):
                tv = cpd.target_vertices_dict[label]
                chem_pot = tv.chem_pot or {}
                comp_phases = ", ".join([str(x) for x in tv.competing_phases]) if tv.competing_phases else "-"
                imp_phases = ", ".join([str(x) for x in tv.impurity_phases]) if tv.impurity_phases else "-"
                mu_parts = [_fmt_mu(el, chem_pot) for el in vertex_elements]
                mu_parts_all = [_fmt_mu(el, chem_pot) for el in all_vertex_elements]
                vertex_entries.append(
                    {
                        "label": label,
                        "chem_pot": chem_pot,
                        "mu_parts": mu_parts,
                        "mu_parts_all": mu_parts_all,
                        "competing": comp_phases,
                        "impurity": imp_phases,
                    }
                )

        def _round_down(x, step):
            return np.floor(x / step) * step

        def _round_up(x, step):
            return np.ceil(x / step) * step
        
        def _close_loop(xs, ys):
            if not xs or not ys:
                return xs, ys
            if xs[0] != xs[-1] or ys[0] != ys[-1]:
                xs = xs + [xs[0]]
                ys = ys + [ys[0]]
            return xs, ys

        if cpd.dim == 2 or project_indices:
            for comp, verts in cpd.polygons.items():
                verts_f = [_as_float_list(v) for v in verts]
                if project_indices:
                    verts_f = [[vv[i] for i in project_indices] for vv in verts_f]
                # sort vertices to avoid self-crossing polygon
                verts_sorted = _sort_coords_2d(verts_f)
                xs = [v[0] for v in verts_sorted]
                ys = [v[1] for v in verts_sorted]
                xs, ys = _close_loop(xs, ys)
                try:
                    atomic_fractions = cpd.atomic_fractions(comp)
                    color = [(f + 1.5) / 2.5 for f in atomic_fractions]
                    rgb = [int(255 * min(max(c, 0.0), 1.0)) for c in color]
                    fill_color = f"rgba({rgb[0]},{rgb[1]},{rgb[2]},1.0)"
                except Exception:
                    fill_color = "rgba(120,120,120,1.0)"
                traces.append(go.Scatter(
                    x=xs, y=ys, mode="lines",
                    line=dict(color="black", width=1),
                    fill="toself",
                    fillcolor=fill_color,
                    name=comp, hoverinfo="skip"
                ))

            # Highlight target stability region if available
            if cpd.target_coords:
                target_coords = [_as_float_list(v) for v in cpd.target_coords.values()]
                if project_indices:
                    target_coords = [[vv[i] for i in project_indices] for vv in target_coords]
                if len(target_coords) >= 3:
                    target_sorted = _sort_coords_2d(target_coords)
                    txs = [v[0] for v in target_sorted]
                    tys = [v[1] for v in target_sorted]
                    txs, tys = _close_loop(txs, tys)
                    traces.append(go.Scatter(
                        x=txs, y=tys, mode="lines",
                        line=dict(color="rgba(30,144,255,0.8)", width=2),
                        fill="toself",
                        fillcolor="rgba(30,144,255,0.18)",
                        name=f"{cpd.target} stability region",
                        hoverinfo="skip"
                    ))
                elif len(target_coords) == 2:
                    txs = [target_coords[0][0], target_coords[1][0]]
                    tys = [target_coords[0][1], target_coords[1][1]]
                    traces.append(go.Scatter(
                        x=txs, y=tys, mode="lines",
                        line=dict(color="rgba(30,144,255,0.9)", width=4),
                        name=f"{cpd.target} stability line",
                        hoverinfo="skip"
                    ))
                # Optional quad fill for binary: B -> (x_min, B_y) -> (x_min, y_min) -> (A_x, y_min) -> A
                if (not project_indices) and cpd.target_coords and "A" in cpd.target_coords and "B" in cpd.target_coords:
                    a = _as_float_list(cpd.target_coords["A"])
                    b = _as_float_list(cpd.target_coords["B"])
                    all_pts = []
                    for verts in cpd.polygons.values():
                        for v in verts:
                            all_pts.append(_as_float_list(v))
                    if all_pts:
                        xs_all = [p[0] for p in all_pts]
                        ys_all = [p[1] for p in all_pts]
                        x_min = min(xs_all)
                        y_min = min(ys_all)
                        quad_x = [a[0], b[0], b[0], a[0], a[0]]
                        quad_y = [a[1], b[1], y_min, y_min, a[1]]
                        traces.append(go.Scatter(
                            x=quad_x, y=quad_y, mode="lines",
                            line=dict(color="rgba(30,144,255,0.8)", width=2),
                            fill="toself",
                            fillcolor="rgba(30,144,255,0.18)",
                            name=f"{cpd.target} stability quad",
                            hoverinfo="skip"
                        ))

            for comp, center in cpd.comp_centers.items():
                center_f = _as_float_list(center)
                if project_indices:
                    center_f = [center_f[i] for i in project_indices]
                traces.append(go.Scatter(
                    x=[center_f[0]], y=[center_f[1]],
                    mode="text", text=[comp],
                    textfont=dict(size=15, color="black"),
                    showlegend=False, hoverinfo="skip"
                ))

            if cpd.target_coords and vertex_entries:
                for entry in vertex_entries:
                    label = entry["label"]
                    coord = cpd.target_coords.get(label)
                    if coord is None:
                        continue
                    coord_f = _as_float_list(coord)
                    if project_indices:
                        coord_f = [coord_f[i] for i in project_indices]
                    hover_text = f"Vertex {label}<br>" + "<br>".join(entry["mu_parts_all"] if project_indices else entry["mu_parts"])
                    traces.append(go.Scatter(
                        x=[coord_f[0]], y=[coord_f[1]],
                        mode="markers+text",
                        text=[label],
                        textposition="top center",
                        textfont=dict(size=15, color="blue"),
                        marker=dict(size=8, color="blue"),
                        showlegend=False,
                        hovertemplate=hover_text + "<extra></extra>"
                    ))

            if project_indices:
                title_used = title + f" (projected: {'-'.join(project_elements_used)})"
                x_title = f"Chemical potential of {project_elements_used[0]}"
                y_title = f"Chemical potential of {project_elements_used[1]}"
            else:
                title_used = title
                x_title = f"Chemical potential of {cpd.vertex_elements[0]}"
                y_title = f"Chemical potential of {cpd.vertex_elements[1]}"

            layout = go.Layout(
                title=title_used,
                autosize=True,
                margin=dict(l=30, r=30, t=40, b=30),
                xaxis=dict(
                    title=x_title,
                    scaleanchor="y",
                    scaleratio=1
                ),
                yaxis=dict(title=y_title)
            )

            # Snap axis ranges to a fixed grid (0.5) for clean boundaries
            if cpd.polygons:
                all_pts = []
                for verts in cpd.polygons.values():
                    for v in verts:
                        vv = _as_float_list(v)
                        if project_indices:
                            vv = [vv[i] for i in project_indices]
                        all_pts.append(vv)
                if all_pts:
                    xs_all = [p[0] for p in all_pts]
                    ys_all = [p[1] for p in all_pts]
                    step = 0.5
                    x_min = _round_down(min(xs_all), step)
                    x_max = _round_up(max(xs_all), step)
                    y_min = _round_down(min(ys_all), step)
                    y_max = _round_up(max(ys_all), step)
                    layout.xaxis.update(range=[x_min, x_max])
                    layout.yaxis.update(range=[y_min, y_max])
            fig = go.Figure(data=traces, layout=layout)

        else:
            # 3D interactive plot (rotatable)
            from vise.util.sort_coords import sort_coords
            for comp, verts in cpd.polygons.items():
                verts_f = [_as_float_list(v) for v in verts]
                verts_sorted = sort_coords(np.array(verts_f)).tolist()
                xs = [v[0] for v in verts_sorted]
                ys = [v[1] for v in verts_sorted]
                zs = [v[2] for v in verts_sorted]

                if len(verts_sorted) >= 3:
                    i = []
                    j = []
                    k = []
                    for idx in range(1, len(verts_sorted) - 1):
                        i.append(0)
                        j.append(idx)
                        k.append(idx + 1)

                    # color by atomic fractions (similar to mpl)
                    try:
                        atomic_fractions = cpd.atomic_fractions(comp)
                        color = [(f + 1.5) / 2.5 for f in atomic_fractions]
                        rgb = [int(255 * min(max(c, 0.0), 1.0)) for c in color]
                        face_color = f"rgba({rgb[0]},{rgb[1]},{rgb[2]},1.0)"
                    except Exception:
                        face_color = "rgba(120,120,120,1.0)"

                    traces.append(go.Mesh3d(
                        x=xs, y=ys, z=zs,
                        i=i, j=j, k=k,
                        color=face_color,
                        opacity=1.0,
                        name=comp,
                        hoverinfo="skip",
                        showscale=False
                    ))

                xs_l = xs + [xs[0]]
                ys_l = ys + [ys[0]]
                zs_l = zs + [zs[0]]
                traces.append(go.Scatter3d(
                    x=xs_l, y=ys_l, z=zs_l,
                    mode="lines",
                    line=dict(color="black", width=2),
                    name=comp,
                    hoverinfo="skip",
                    showlegend=False
                ))

            for comp, center in cpd.comp_centers.items():
                center_f = _as_float_list(center)
                traces.append(go.Scatter3d(
                    x=[center_f[0]], y=[center_f[1]], z=[center_f[2]],
                    mode="text", text=[comp],
                    textfont=dict(size=15, color="black"),
                    showlegend=False, hoverinfo="skip"
                ))

            if cpd.target_coords and vertex_entries:
                for entry in vertex_entries:
                    label = entry["label"]
                    coord = cpd.target_coords.get(label)
                    if coord is None:
                        continue
                    coord_f = _as_float_list(coord)
                    hover_text = f"Vertex {label}<br>" + "<br>".join(entry["mu_parts"])
                    traces.append(go.Scatter3d(
                        x=[coord_f[0]], y=[coord_f[1]], z=[coord_f[2]],
                        mode="markers+text",
                        text=[label],
                        textposition="top center",
                        textfont=dict(size=15, color="blue"),
                        marker=dict(size=4, color="blue"),
                        showlegend=False,
                        hovertemplate=hover_text + "<extra></extra>"
                    ))

            layout = go.Layout(
                title=title,
                autosize=True,
                margin=dict(l=0, r=0, t=40, b=0),
                scene=dict(
                    aspectmode="data",
                    xaxis=dict(title=f"Chemical potential of {cpd.vertex_elements[0]}"),
                    yaxis=dict(title=f"Chemical potential of {cpd.vertex_elements[1]}"),
                    zaxis=dict(title=f"Chemical potential of {cpd.vertex_elements[2]}"),
                ),
                showlegend=False
            )
            fig = go.Figure(data=traces, layout=layout)

        target_formula = getattr(cpd, "target", None) or "N/A"
        vertex_info_lines = []
        if vertex_entries:
            for entry in vertex_entries:
                mu_str = ", ".join(entry["mu_parts"]) if entry["mu_parts"] else "-"
                vertex_info_lines.append(
                    "<li>"
                    f"<b>{entry['label']}</b>"
                    "<div class=\"vertex-meta\">"
                    f"<div><span class=\"k\">μ</span>: {mu_str}</div>"
                    f"<div><span class=\"k\">competing</span>: {entry['competing']}</div>"
                    f"<div><span class=\"k\">impurity</span>: {entry['impurity']}</div>"
                    "</div>"
                    "</li>"
                )
        vertex_html = "<ul>" + "".join(vertex_info_lines) + "</ul>" if vertex_info_lines else "<p>-</p>"

        plot_div = fig.to_html(include_plotlyjs="cdn", full_html=False, config={"responsive": True, "scrollZoom": True})
        is_2d_plot = cpd.dim == 2 or project_indices
        html = f"""<!doctype html>
<html>
  <head>
    <meta charset="utf-8">
    <title>CPD</title>
    <style>
      html, body {{ height: 100%; margin: 0; }}
      body {{ display: grid; grid-template-columns: 1fr minmax(280px, 420px); font-family: Arial, sans-serif; }}
      #plot {{ width: 100%; height: 100vh; display: flex; align-items: center; justify-content: center; overflow: hidden; }}
      #plot > div {{ width: 100%; height: 100%; }}
      #info {{
        padding: 16px 18px;
        box-sizing: border-box;
        border-left: 1px solid #ddd;
        background: #fafafa;
        overflow-y: auto;
        z-index: 10;
      }}
      h3 {{ margin: 0 0 8px 0; }}
      .section {{ margin-bottom: 14px; }}
      ul {{ padding-left: 18px; margin: 6px 0; }}
      .vertex-meta {{ margin: 4px 0 8px 0; font-size: 0.95em; }}
      .vertex-meta .k {{ font-weight: 600; }}
      .btn-row {{ display: flex; gap: 8px; flex-wrap: wrap; margin-top: 6px; }}
      .btn-row button {{
        padding: 6px 10px;
        border: 1px solid #bbb;
        background: #fff;
        border-radius: 4px;
        cursor: pointer;
      }}
      .btn-row button:disabled {{
        opacity: 0.5;
        cursor: not-allowed;
      }}
    </style>
  </head>
  <body>
    <div id="plot">{plot_div}</div>
    <div id="info">
      <div class="section"><h3>CPD Info</h3></div>
      <div class="section"><b>(1)</b> Chemical formula: {target_formula}</div>
      <div class="section">
        <b>Rotate (2D only)</b>
        <div class="btn-row">
          <button id="rotate-cw">Rotate 90° CW</button>
          <button id="rotate-ccw">Rotate 90° CCW</button>
        </div>
      </div>
      <div class="section"><b>(2)</b> Vertices A, B, C, D: competing phases / impurity phases
        {vertex_html}
      </div>
    </div>
    <script>
      (function() {{
        const is2d = {str(is_2d_plot).lower()};
        const btnCw = document.getElementById('rotate-cw');
        const btnCcw = document.getElementById('rotate-ccw');
        if (!is2d) {{
          btnCw.disabled = true;
          btnCcw.disabled = true;
          return;
        }}
        const plot = document.querySelector('#plot .plotly-graph-div');
        if (!plot) return;
        let swapped = false;
        const origXT = plot.layout?.xaxis?.title?.text || plot.layout?.xaxis?.title || '';
        const origYT = plot.layout?.yaxis?.title?.text || plot.layout?.yaxis?.title || '';

        function toArray(v) {{
          if (Array.isArray(v)) return v.slice();
          return [v];
        }}

        function computeRange(traces) {{
          let xs = [];
          let ys = [];
          traces.forEach(t => {{
            if (t.x && t.y) {{
              xs = xs.concat(toArray(t.x));
              ys = ys.concat(toArray(t.y));
            }}
          }});
          const xMin = Math.min.apply(null, xs);
          const xMax = Math.max.apply(null, xs);
          const yMin = Math.min.apply(null, ys);
          const yMax = Math.max.apply(null, ys);
          return {{ xMin, xMax, yMin, yMax }};
        }}

        function rotate(dir) {{
          const newData = plot.data.map(t => {{
            if (!t.x || !t.y) return t;
            const x = toArray(t.x);
            const y = toArray(t.y);
            let nx, ny;
            if (dir === 'cw') {{
              nx = y;
              ny = x.map(v => -v);
            }} else {{
              nx = y.map(v => -v);
              ny = x;
            }}
            return {{ ...t, x: nx, y: ny }};
          }});

          const r = computeRange(newData);
          const newLayout = {{
            ...plot.layout,
            xaxis: {{ ...plot.layout.xaxis, range: [r.xMin, r.xMax], scaleanchor: 'y', scaleratio: 1 }},
            yaxis: {{ ...plot.layout.yaxis, range: [r.yMin, r.yMax] }}
          }};

          if (!swapped) {{
            newLayout.xaxis.title = {{ text: origYT }};
            newLayout.yaxis.title = {{ text: origXT }};
          }} else {{
            newLayout.xaxis.title = {{ text: origXT }};
            newLayout.yaxis.title = {{ text: origYT }};
          }}
          swapped = !swapped;
          Plotly.react(plot, newData, newLayout, {{ responsive: true, scrollZoom: true }});
        }}

        btnCw.addEventListener('click', () => rotate('cw'));
        btnCcw.addEventListener('click', () => rotate('ccw'));
      }})();
    </script>
  </body>
</html>
"""
        Path(f"{output_prefix}.html").write_text(html)
        print(f"Generated {output_prefix}.html")

    return True


def maybe_plot_cpd_pdf(cpd_json_path="chem_pot_diag.json"):
    cpd_path = Path(cpd_json_path)
    if not cpd_path.exists():
        print(f"chem_pot_diag.json not found at {cpd_path}. Skipping cpd.pdf.")
        return False

    try:
        cpd = loadfn(cpd_path)
        return plot_cpd_outputs(cpd, output_prefix="cpd", make_pdf=True, make_html=True)
    except Exception as e:
        print(f"Failed to generate cpd.pdf: {e}")
        return False


def find_interstitial_sites(input_path: Path,
                            output_dir: Path = Path("."),
                            min_dist: float = 0.5,
                            insert_species: str = "H"):
    try:
        from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
    except ImportError:
        print("Error: pymatgen-analysis-defects package is required for Voronoi interstitial finding.")
        print("Please install it via: pip install pymatgen-analysis-defects")
        return None

    from pydefect.input_maker.local_extrema import VolumetricDataLocalExtrema, CoordInfo, VolumetricDataAnalyzeParams
    from pydefect.util.structure_tools import Distances

    struct = Structure.from_file(str(input_path))
    gen = VoronoiInterstitialGenerator(min_dist=min_dist)
    interstitials = list(gen.generate(struct, insert_species=[insert_species]))

    if not interstitials:
        print("No interstitials found! Try decreasing min_dist.")
        return None

    extrema_points = []
    for i, defect in enumerate(interstitials, 1):
        site = defect.site
        coords = [float(x) for x in site.frac_coords]
        try:
            site_sym = defect.site_symmetry
        except Exception:
            site_sym = "unknown"

        coordination = Distances(struct, coords).coordination()
        coord_info = CoordInfo(
            site_symmetry=site_sym,
            coordination=coordination,
            frac_coords=[coords],
            quantities=[0.0]
        )
        extrema_points.append(coord_info)

        mult = getattr(defect, "multiplicity", None)
        mult_str = f", mult={mult}" if mult is not None else ""
        print(f"{i:>3} | {coords}{mult_str}")

    params = VolumetricDataAnalyzeParams(
        threshold_frac=None,
        threshold_abs=None,
        min_dist=float(min_dist),
        tol=0.0,
        radius=0.0
    )
    local_extrema = VolumetricDataLocalExtrema(
        unit_cell=struct,
        is_min=True,
        extrema_points=extrema_points,
        info=f"Voronoi_interstitials({insert_species})",
        params=params
    )

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    stem = input_path.name
    if stem.startswith("POSCAR"):
        prefix = stem
    else:
        prefix = input_path.stem
    out_path = output_dir / f"interstitial-site-{prefix}.json"
    local_extrema.to_json_file(str(out_path))
    print(f"Wrote {out_path}")
    return out_path

def stabilize_target(relative_energies_path, target_formula, manual_shift=0.0, buffer=1e-5):
    """
    Checks if the target is above convex hull. If so, shifts its energy in relative_energies.yaml
    to make it stable. Returns the total applied shift value (0.0 if stable and no manual shift).
    """
    if not Path(relative_energies_path).exists():
        print(f"Error: {relative_energies_path} does not exist.")
        return 0.0

    with open(relative_energies_path, 'r') as f:
        rel_energies = yaml.safe_load(f)

    # 1. Apply manual shift first (temporarily to check stability or permanently if stable)
    if target_formula not in rel_energies:
        print(f"Error: Target {target_formula} not found in {relative_energies_path}")
        return 0.0
    
    original_energy = rel_energies[target_formula]
    # We apply the manual shift (subtracted, as per convention: positive shift means lowering energy? 
    # The user said "--energy-shift-target 0.05 이면 target energy 를 0.05 eV 낮춤" -> new = old - 0.05)
    # So if manual_shift is 0.05, we subtract 0.05.
    current_energy_after_manual = original_energy - manual_shift
    rel_energies[target_formula] = current_energy_after_manual
    
    entries = []
    target_entry = None

    # Re-build entries with the manual shift applied
    for formula, energy_per_atom in rel_energies.items():
        try:
            comp = Composition(formula)
            total_energy = energy_per_atom * comp.num_atoms
            entry = PDEntry(comp, total_energy)
            entries.append(entry)
            if formula == target_formula:
                target_entry = entry
        except Exception as e:
            print(f"Warning: Could not process formula {formula}: {e}")

    if target_entry is None:
        return 0.0

    entries = []
    target_entry = None
    target_formula_norm = Composition(target_formula).reduced_formula

    # Re-build entries with the manual shift applied and normalized formulas
    for formula, energy_per_atom in rel_energies.items():
        try:
            comp = Composition(formula)
            norm_f = comp.reduced_formula
            total_energy = energy_per_atom * comp.num_atoms
            # Use normalized formula as the name for consistent matching
            entry = PDEntry(comp, total_energy, name=norm_f)
            entries.append(entry)
            
            if norm_f == target_formula_norm and formula == target_formula:
                target_entry = entry
        except Exception as e:
            print(f"Warning: Could not process formula {formula}: {e}")

    if target_entry is None:
        print(f"Error: Target entry for {target_formula} could not be created.")
        return 0.0

    pd = PhaseDiagram(entries)
    e_above_hull = pd.get_e_above_hull(target_entry)
    
    # --- Generate convex_hull_energy.yaml ---
    try:
        # Use a small threshold to determine stability
        is_stable = e_above_hull <= 1e-8
        
        # 1. Standard stability info (including target in PD)
        if is_stable:
            competing_phases_info = {target_formula: 1.0}
            convex_hull_energy_per_atom = float(target_entry.energy_per_atom)
        else:
            decomp = pd.get_decomposition(target_entry.composition)
            convex_hull_energy_per_atom = target_entry.energy_per_atom - e_above_hull
            competing_phases_info = {entry.name: float(amount) for entry, amount in decomp.items()}

        # 2. Benchmark stability info (excluding target from PD)
        decomposition_info = None
        try:
            entries_excl_target = [e for e in entries if e.composition.reduced_formula != target_formula_norm]
            if entries_excl_target:
                pd_benchmark = PhaseDiagram(entries_excl_target)
                hull_energy_total = pd_benchmark.get_hull_energy(target_entry.composition)
                hull_energy_per_atom = hull_energy_total / target_entry.composition.num_atoms
                e_above_hull_benchmark = target_entry.energy_per_atom - hull_energy_per_atom
                
                comp_phases_bench = {}
                try:
                    decomp_benchmark = pd_benchmark.get_decomposition(target_entry.composition)
                    comp_phases_bench = {entry.name: float(amount) for entry, amount in decomp_benchmark.items()}
                except Exception: pass
                
                decomposition_info = {
                    "hull_distance": float(e_above_hull_benchmark),
                    "competing_phases": comp_phases_bench
                }
        except Exception as e_bench:
            print(f"Warning: Benchmark stability calculation failed: {e_bench}")

        # Construct dictionary in specific order
        convex_hull_data = {
            "target": target_formula,
            "formation_energy_per_atom": float(target_entry.energy_per_atom),
            "convex_hull_energy_per_atom": float(convex_hull_energy_per_atom),
            "e_above_hull": float(e_above_hull)
        }
        if decomposition_info:
            convex_hull_data["decomposition"] = decomposition_info

        # Write to yaml in the same directory as relative_energies.yaml
        output_dir = Path(relative_energies_path).parent
        with open(output_dir / "convex_hull_energy.yaml", "w") as f:
            yaml.dump(convex_hull_data, f, default_flow_style=False)
        print(f"Generated convex_hull_energy.yaml")
        
    except Exception as e:
        print(f"Warning: Failed to generate convex_hull_energy.yaml details: {e} (e_h: {e_above_hull})")
    # ----------------------------------------
    # ----------------------------------------

    extra_stabilization_shift = 0.0

    if e_above_hull > 1e-8:  # Use a small epsilon
        # It's still unstable after manual shift (or manual shift was 0)
        required_shift = -1.0 * e_above_hull
        # Add buffer
        extra_stabilization_shift = required_shift - buffer
        
        print(f"Target {target_formula} is unstable (e_above_hull={e_above_hull:.4f} eV/atom).")
        print(f"Additional stabilization shift required: {extra_stabilization_shift:.6f} eV/atom.")

    # Total shift = manual_shift (already subtracted) + extra_stabilization_shift (negative value to lower energy further)
    # But wait, user said "shift 0.05 means lower by 0.05".
    # My stabilize logic was "return recommended_shift" where recommended_shift was negative.
    # So let's keep consistent: return value is the net change in energy.
    # manual_shift argument is positive -> net change is -manual_shift.
    # extra_stabilization_shift is negative.
    
    total_change = -manual_shift + extra_stabilization_shift
    
    if total_change != 0.0:
        final_energy = float(original_energy + total_change)
        rel_energies[target_formula] = final_energy
        
        with open(relative_energies_path, 'w') as f:
            yaml.dump(rel_energies, f)
            
        print(f"Total applied energy shift: {total_change:.6f} eV/atom (Manual: -{manual_shift}, Stabilization: {extra_stabilization_shift:.6f}).")
        return total_change

    return 0.0

def write_summary_at_vertices(summary_path: Path, applied_shift=0.0):
    """
    Calculates and prints defect formation energies for each chemical potential
    vertex from a defect_energy_summary.json file.
    """
    if not summary_path.exists():
        print(f"Error: {summary_path} not found.")
        return

    try:
        # loadfn will deserialize into DefectEnergySummary object
        summary = loadfn(summary_path)
    except Exception as e:
        print(f"Error loading {summary_path}: {e}")
        return

    rel_chem_pots = summary.rel_chem_pots
    defect_energies_summary = summary.defect_energies
    output_filename = "defect_energy_summary-at-vertices.txt"

    with open(output_filename, "w") as f:
        f.write(f"title: {summary.title}\n")
        if applied_shift != 0.0:
            f.write(f"# Energy shift applied to make target stable: {applied_shift} eV/atom\n")
        f.write("\n")

        f.write("--- Standard Formation Energies (Chemical Potential = 0) ---\n")
        f.write(f"{ 'Defect Name':<20} {'Charge':<10} {'Formation Energy (eV)'}\n")
        f.write(f"{'-'*20} {'-'*10} {'-'*25}\n")
        
        for defect_name, defect_group in defect_energies_summary.items():
             if 0 in defect_group.charges:
                 idx = defect_group.charges.index(0)
                 energy = defect_group.defect_energies[idx].formation_energy
                 f.write(f"{defect_name:<20} {'0':<10} {energy:>25.4f}\n")
        f.write("\n")

        for vertex_label, chem_pots in rel_chem_pots.items():
            f.write(f"--- Vertex: {vertex_label} ---\n")
            chem_pot_str = ", ".join([f"{elem}: {pot:.4f}" for elem, pot in chem_pots.items()])
            f.write(f"Relative chemical potentials: {chem_pot_str}\n")
            f.write("-" * 40 + "\n")
            f.write(f"{ 'Defect Name':<20} {'Formation Energy (eV)'}\n")
            f.write(f"{'-'*20} {'-'*25}\n")

            formation_energies = {}
            for defect_name, defect_group in defect_energies_summary.items():
                charge_0_defect_energy = None
                try:
                    if 0 in defect_group.charges:
                        charge_0_index = defect_group.charges.index(0)
                        charge_0_defect_energy = defect_group.defect_energies[charge_0_index]
                except (ValueError, AttributeError):
                    pass

                if not charge_0_defect_energy:
                    continue

                atom_io = defect_group.atom_io
                base_energy = charge_0_defect_energy.formation_energy
                
                reservoir_contribution = 0.0
                for element, num_diff in atom_io.items():
                    if element in chem_pots:
                        reservoir_contribution -= num_diff * chem_pots[element]

                formation_energy = base_energy + reservoir_contribution
                formation_energies[defect_name] = formation_energy
            
            sorted_defects = sorted(formation_energies.items(), key=lambda item: item[0])
            for name, energy in sorted_defects:
                f.write(f"{name:<20} {energy:>25.4f}\n")
            
            f.write("\n")

    print(f"Output written to {output_filename}")


def _postprocess_defect_artifacts(d: Path, supercell_info):
    if not supercell_info:
        return
    try:
        if Path(d / "CONTCAR").exists() and Path(d / "defect_entry.json").exists():
            final_structure = Structure.from_file(d / "CONTCAR")
            defect_entry = loadfn(d / "defect_entry.json")
            dsi_maker = MakeDefectStructureInfo(
                perfect=supercell_info.structure,
                initial=defect_entry.structure,
                final=final_structure,
                symprec=defaults.symmetry_length_tolerance,
                dist_tol=defaults.dist_tol
            )
            dsi_maker.defect_structure_info.to_json_file(d / "defect_structure_info.json")
            print(f"Generated defect_structure_info.json for {d.name}")
            try:
                dsi_loaded = loadfn(d / "defect_structure_info.json")
                vesta_maker = MakeDefectVestaFile(dsi_loaded)
                vesta_maker.initial_vesta.write_file(d / "defect_initial.vesta")
                vesta_maker.final_vesta.write_file(d / "defect.vesta")
                print(f"Generated VESTA files for {d.name}")
            except Exception as e_vesta:
                print(f"Warning: Failed to generate VESTA files: {e_vesta}")
    except Exception as e:
        print(f"Warning: Failed to generate defect_structure_info.json for {d.name}: {e}")


def _relax_single_dir_worker(params):
    d, ff, model_path, device, fmax, isif, modal, verbose = params
    d = Path(d)
    try:
        calc_kwargs = {"model_path": model_path, "device": device, "modal": modal}
        if ff == "mace":
            calc_kwargs["model_paths"] = [calc_kwargs["model_path"]]
            del calc_kwargs["model_path"]
        calculator = get_calculator(ff_name=ff, **calc_kwargs)

        cwd = Path.cwd()
        os.chdir(d)
        try:
            if verbose:
                print(f"Relaxing in {d.name}...")
            with Logger("relax-POSCAR_log.txt") as lg:
                original_stdout = sys.stdout
                try:
                    sys.stdout = lg
                    write_pydefect_dummy_files(".")
                    relax_structure(
                        input_file="POSCAR",
                        fmax=fmax,
                        isif=isif,
                        device=device,
                        calculator=calculator,
                        ff=ff,
                        outcar_name="OUTCAR",
                        contcar_name="CONTCAR",
                        xml_name="vasprun.xml",
                        make_pdf=True,
                        write_json=True,
                        modal=modal
                    )
                finally:
                    sys.stdout = original_stdout
        finally:
            os.chdir(cwd)

        ok = (d / "vasprun.xml").exists() or (d / "OUTCAR").exists()
        return str(d), ok, None
    except Exception as e:
        return str(d), False, str(e)


def run_macer_relax(target_dirs, isif=3, supercell_info=None, ff=None, model_path=None, device=None, fmax=0.03, verbose=False, modal=None, n_jobs=1):
    """
    Unified relaxation function for CPD and Defect workflows.
    Iterates over target_dirs and runs macer relaxation.
    """
    if not target_dirs:
        return []

    # Determine FF and Model if not provided
    if ff is None:
        available_ffs = get_available_ffs()
        if not available_ffs:
            print("Error: No MLFF packages installed.")
            return []
        ff = available_ffs[0]
    
    if device is None:
        device = DEFAULT_DEVICE
    
    if model_path is None:
        default_model_name = DEFAULT_MODELS.get(ff)
        if default_model_name:
            FFS_USING_MODEL_NAME = {"fairchem", "orb", "chgnet", "m3gnet"}
            if ff in FFS_USING_MODEL_NAME:
                model_path = default_model_name
            else:
                model_path = resolve_model_path(default_model_name)
    else:
        model_path = resolve_model_path(model_path)
    
    if n_jobs < 1:
        raise ValueError("Error: --n-jobs must be >= 1.")

    print(f"Initializing {ff.upper()} calculator on {device}...")
    successful_dirs = []
    use_parallel = n_jobs > 1 and device == "cpu"
    if n_jobs > 1 and device != "cpu":
        print(f"WARNING: --n-jobs>1 is supported only with --device cpu (current: {device}). Falling back to serial.")

    if use_parallel:
        print(f"Using parallel relaxation with {n_jobs} workers.")
        params = [
            (str(Path(d).resolve()), ff, model_path, device, fmax, isif, modal, verbose)
            for d in target_dirs
        ]
        with ProcessPoolExecutor(max_workers=n_jobs) as pool:
            for d_str, ok, err in pool.map(_relax_single_dir_worker, params):
                d = Path(d_str)
                if ok:
                    successful_dirs.append(d)
                    _postprocess_defect_artifacts(d, supercell_info)
                else:
                    print(f"Relaxation failed in {d.name}: {err}")
        return successful_dirs

    calc_kwargs = {"model_path": model_path, "device": device, "modal": modal}
    if ff == "mace":
        calc_kwargs["model_paths"] = [calc_kwargs["model_path"]]
        del calc_kwargs["model_path"]

    try:
        calculator = get_calculator(ff_name=ff, **calc_kwargs)
        actual_path = None
        if hasattr(calculator, "model_path"):
            actual_path = calculator.model_path
        elif hasattr(calculator, "model_paths") and calculator.model_paths:
            actual_path = calculator.model_paths[0]
        if actual_path:
            print(f"  MLFF Model: {actual_path}")
        print("Calculator initialized.")
    except Exception as e:
        print(f"Failed to initialize calculator: {e}")
        return []

    cwd = Path.cwd()
    original_stdout = sys.stdout
    for d in target_dirs:
        if verbose:
            print(f"Relaxing in {d.name}...")
        os.chdir(d)
        try:
            with Logger("relax-POSCAR_log.txt") as lg:
                sys.stdout = lg
                write_pydefect_dummy_files(".")
                relax_structure(
                    input_file="POSCAR",
                    fmax=fmax,
                    isif=isif,
                    device=device,
                    calculator=calculator,
                    ff=ff,
                    outcar_name="OUTCAR",
                    contcar_name="CONTCAR",
                    xml_name="vasprun.xml",
                    make_pdf=True,
                    write_json=True,
                    modal=modal
                )
            if Path("vasprun.xml").exists() or Path("OUTCAR").exists():
                successful_dirs.append(d)
                _postprocess_defect_artifacts(d, supercell_info)
        except Exception as e:
            sys.stdout = original_stdout
            print(f"Relaxation failed in {d.name}: {e}")
        finally:
            sys.stdout = original_stdout
            os.chdir(cwd)

    return successful_dirs

def run_voronoi_interstitial_finding(supercell_info_path: Path):
    """
    Finds interstitial sites using Voronoi tessellation via pymatgen-analysis-defects,
    and appends unique symmetry-independent sites to supercell_info.json.
    """
    if not supercell_info_path.exists():
        print(f"Error: {supercell_info_path} not found.")
        return

    print(f"Loading {supercell_info_path} for interstitial finding...")
    
    try:
        from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
    except ImportError:
        print("Error: pymatgen-analysis-defects package is required for Voronoi interstitial finding.")
        print("Please install it via: pip install pymatgen-analysis-defects")
        return

    try:
        from pydefect.input_maker.append_interstitial import append_interstitial
    except ImportError as e:
        print(f"Required package not found: {e}")
        return

    try:
        supercell_info = loadfn(supercell_info_path)
        unitcell = supercell_info.unitcell_structure
        # Create structure from unitcell dict
        struct = Structure.from_dict(unitcell.as_dict())

        print(f"Unitcell loaded: {struct.composition.reduced_formula}")
        
        # Parameters
        min_dist = 0.5
        insert_species = "H"
        
        print(f"Generating Voronoi interstitial candidates (min_dist={min_dist}, species={insert_species})...")
        
        gen = VoronoiInterstitialGenerator(min_dist=min_dist)
        # generate() returns a generator, convert to list
        interstitials = list(gen.generate(struct, insert_species=[insert_species]))
        
        if not interstitials:
            print("No interstitials found! Try decreasing min_dist.")
            return

        print(f"Found {len(interstitials)} distinct interstitial sites.")

        coords_list = []
        infos = []

        print("\n--- Sites to Add ---")
        header = f"{ 'Idx':<4} | {'Fractional Coords':<30} | {'Mult':<4}"
        print(header)
        print("-" * len(header))

        for i, defect in enumerate(interstitials):
            site = defect.site
            coords = site.frac_coords.tolist()
            mult = defect.multiplicity
            
            info_str = f"Voronoi_site_{i}_mult_{mult}"
            
            print(f"{i:<4} | {str(coords):<30} | {mult:<4}")
            
            coords_list.append(coords)
            infos.append(info_str)

        print("-" * len(header))

        print("\nAppending interstitials to SupercellInfo...")
        new_supercell_info = append_interstitial(
            supercell_info=supercell_info,
            unitcell_structure=unitcell,
            frac_coords=coords_list,
            infos=infos
        )
        
        new_supercell_info.to_json_file(supercell_info_path)
        print(f"Successfully updated {supercell_info_path} with {len(interstitials)} interstitial types.")

    except Exception as e:
        print(f"Error during interstitial finding: {e}")
        import traceback
        traceback.print_exc()

def write_simple_defect_summary(root_dir: Path):
    """
    Generates a simplified defect energy summary from individual defect_energy_info.json files.
    Used when target_vertices.yaml is not available.
    """
    root_dir = Path(root_dir)
    defect_energies = []

    print(f"Scanning for defect_energy_info.yaml files in {root_dir.absolute()}...")

    for d in root_dir.iterdir():
        if d.is_dir():
            info_path = d / "defect_energy_info.yaml"
            if info_path.exists():
                try:
                    info = loadfn(info_path)
                    # When loaded, it can be a dict-like object (CommentedMap)
                    name = info['name']
                    charge = info['charge']
                    formation_energy = info['formation_energy']
                    defect_energies.append((name, charge, formation_energy))
                except Exception as e:
                    print(f"Warning: Failed to load or parse {info_path}: {e}")

    if not defect_energies:
        print("No defect energy info found.")
        return

    # Sort by name
    defect_energies.sort(key=lambda x: x[0])

    output_filename = "defect_energy_summary-at-vertices.txt"
    with open(output_filename, "w") as f:
        f.write("title: Simplified Defect Summary\n\n")
        f.write("--- Standard Formation Energies (Chemical Potential = 0) ---\n")
        f.write(f"{ 'Defect Name':<20} {'Charge':<10} {'Formation Energy (eV)'}\n")
        f.write(f"{'-'*20} {'-'*10} {'-'*25}\n")
        
        for name, charge, energy in defect_energies:
             f.write(f"{name:<20} {charge:<10} {energy:>25.4f}\n")
    
    print(f"Output written to {output_filename}")



###########################################################################################
# Implementation of different loss functions
# Authors: Ilyes Batatia, Gregor Simm
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

from typing import Optional

import torch
import torch.distributed as dist

from macer.externals.mace_bundled.tools import TensorDict
from macer.externals.mace_bundled.tools.torch_geometric import Batch


# ------------------------------------------------------------------------------
# Helper function for loss reduction that handles DDP correction
# ------------------------------------------------------------------------------
def is_ddp_enabled():
    return dist.is_initialized() and dist.get_world_size() > 1


def reduce_loss(raw_loss: torch.Tensor, ddp: Optional[bool] = None) -> torch.Tensor:
    """
    Reduces an element-wise loss tensor.

    If ddp is True and distributed is initialized, the function computes:

        loss = (local_sum * world_size) / global_num_elements

    Otherwise, it returns the regular mean.
    """
    ddp = is_ddp_enabled() if ddp is None else ddp
    if ddp and dist.is_initialized():
        world_size = dist.get_world_size()
        n_local = raw_loss.numel()
        loss_sum = raw_loss.sum()
        total_samples = torch.tensor(
            n_local, device=raw_loss.device, dtype=raw_loss.dtype
        )
        dist.all_reduce(total_samples, op=dist.ReduceOp.SUM)
        return loss_sum * world_size / total_samples
    return raw_loss.mean()


# ------------------------------------------------------------------------------
# Energy Loss Functions
# ------------------------------------------------------------------------------


def mean_squared_error_energy(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    raw_loss = torch.square(ref["energy"] - pred["energy"])
    return reduce_loss(raw_loss, ddp)


def weighted_mean_squared_error_energy(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    # Calculate per-graph number of atoms.
    num_atoms = ref.ptr[1:] - ref.ptr[:-1]  # shape: [n_graphs]
    raw_loss = (
        ref.weight
        * ref.energy_weight
        * torch.square((ref["energy"] - pred["energy"]) / num_atoms)
    )
    return reduce_loss(raw_loss, ddp)


def weighted_mean_absolute_error_energy(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    num_atoms = ref.ptr[1:] - ref.ptr[:-1]
    raw_loss = (
        ref.weight
        * ref.energy_weight
        * torch.abs((ref["energy"] - pred["energy"]) / num_atoms)
    )
    return reduce_loss(raw_loss, ddp)


# ------------------------------------------------------------------------------
# Stress and Virials Loss Functions
# ------------------------------------------------------------------------------


def weighted_mean_squared_stress(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    configs_weight = ref.weight.view(-1, 1, 1)
    configs_stress_weight = ref.stress_weight.view(-1, 1, 1)
    raw_loss = (
        configs_weight
        * configs_stress_weight
        * torch.square(ref["stress"] - pred["stress"])
    )
    return reduce_loss(raw_loss, ddp)


def weighted_mean_squared_virials(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    configs_weight = ref.weight.view(-1, 1, 1)
    configs_virials_weight = ref.virials_weight.view(-1, 1, 1)
    num_atoms = (ref.ptr[1:] - ref.ptr[:-1]).view(-1, 1, 1)
    raw_loss = (
        configs_weight
        * configs_virials_weight
        * torch.square((ref["virials"] - pred["virials"]) / num_atoms)
    )
    return reduce_loss(raw_loss, ddp)


# ------------------------------------------------------------------------------
# Forces Loss Functions
# ------------------------------------------------------------------------------


def mean_squared_error_forces(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    # Repeat per-graph weights to per-atom level.
    configs_weight = torch.repeat_interleave(
        ref.weight, ref.ptr[1:] - ref.ptr[:-1]
    ).unsqueeze(-1)
    configs_forces_weight = torch.repeat_interleave(
        ref.forces_weight, ref.ptr[1:] - ref.ptr[:-1]
    ).unsqueeze(-1)
    raw_loss = (
        configs_weight
        * configs_forces_weight
        * torch.square(ref["forces"] - pred["forces"])
    )
    return reduce_loss(raw_loss, ddp)


def mean_normed_error_forces(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    raw_loss = torch.linalg.vector_norm(ref["forces"] - pred["forces"], ord=2, dim=-1)
    return reduce_loss(raw_loss, ddp)


# ------------------------------------------------------------------------------
# Dipole Loss Function
# ------------------------------------------------------------------------------


def weighted_mean_squared_error_dipole(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    num_atoms = (ref.ptr[1:] - ref.ptr[:-1]).unsqueeze(-1)
    raw_loss = torch.square((ref["dipole"] - pred["dipole"]) / num_atoms)
    return reduce_loss(raw_loss, ddp)


# ------------------------------------------------------------------------------
# Polarizability Loss Function
# ------------------------------------------------------------------------------


def weighted_mean_squared_error_polarizability(
    ref: Batch,
    pred: TensorDict,
    ddp: Optional[
        bool
    ] = None,  # ,mean: Optional[torch.Tensor] = None , std: Optional[torch.Tensor] = None
) -> torch.Tensor:
    # polarizability: [n_graphs, ]
    # ref_polar = ref["polarizability"].view(-1, 3, 3) * std.view(1, 3, 3) + mean.view(1, 3, 3) if mean is not None and std is not None else ref["polarizability"]
    num_atoms = (ref.ptr[1:] - ref.ptr[:-1]).view(-1, 1, 1)  # [n_graphs,1]
    raw_loss = torch.square(
        (ref["polarizability"].view(-1, 3, 3) - pred["polarizability"]) / num_atoms
    )
    return reduce_loss(raw_loss, ddp)


# ------------------------------------------------------------------------------
# Conditional Losses for Forces
# ------------------------------------------------------------------------------


def conditional_mse_forces(
    ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
) -> torch.Tensor:
    configs_weight = torch.repeat_interleave(
        ref.weight, ref.ptr[1:] - ref.ptr[:-1]
    ).unsqueeze(-1)
    configs_forces_weight = torch.repeat_interleave(
        ref.forces_weight, ref.ptr[1:] - ref.ptr[:-1]
    ).unsqueeze(-1)
    # Define multiplication factors for different regimes.
    factors = torch.tensor(
        [1.0, 0.7, 0.4, 0.1], device=ref["forces"].device, dtype=ref["forces"].dtype
    )
    err = ref["forces"] - pred["forces"]
    se = torch.zeros_like(err)
    norm_forces = torch.norm(ref["forces"], dim=-1)
    c1 = norm_forces < 100
    c2 = (norm_forces >= 100) & (norm_forces < 200)
    c3 = (norm_forces >= 200) & (norm_forces < 300)
    se[c1] = torch.square(err[c1]) * factors[0]
    se[c2] = torch.square(err[c2]) * factors[1]
    se[c3] = torch.square(err[c3]) * factors[2]
    se[~(c1 | c2 | c3)] = torch.square(err[~(c1 | c2 | c3)]) * factors[3]
    raw_loss = configs_weight * configs_forces_weight * se
    return reduce_loss(raw_loss, ddp)


def conditional_huber_forces(
    ref_forces: torch.Tensor,
    pred_forces: torch.Tensor,
    huber_delta: float,
    ddp: Optional[bool] = None,
) -> torch.Tensor:
    factors = huber_delta * torch.tensor(
        [1.0, 0.7, 0.4, 0.1], device=ref_forces.device, dtype=ref_forces.dtype
    )
    norm_forces = torch.norm(ref_forces, dim=-1)
    c1 = norm_forces < 100
    c2 = (norm_forces >= 100) & (norm_forces < 200)
    c3 = (norm_forces >= 200) & (norm_forces < 300)
    c4 = ~(c1 | c2 | c3)
    se = torch.zeros_like(pred_forces)
    se[c1] = torch.nn.functional.huber_loss(
        ref_forces[c1], pred_forces[c1], reduction="none", delta=factors[0]
    )
    se[c2] = torch.nn.functional.huber_loss(
        ref_forces[c2], pred_forces[c2], reduction="none", delta=factors[1]
    )
    se[c3] = torch.nn.functional.huber_loss(
        ref_forces[c3], pred_forces[c3], reduction="none", delta=factors[2]
    )
    se[c4] = torch.nn.functional.huber_loss(
        ref_forces[c4], pred_forces[c4], reduction="none", delta=factors[3]
    )
    return reduce_loss(se, ddp)


# ------------------------------------------------------------------------------
# Loss Modules Combining Multiple Quantities
# ------------------------------------------------------------------------------


class WeightedEnergyForcesLoss(torch.nn.Module):
    def __init__(self, energy_weight=1.0, forces_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_energy = weighted_mean_squared_error_energy(ref, pred, ddp)
        loss_forces = mean_squared_error_forces(ref, pred, ddp)
        return self.energy_weight * loss_energy + self.forces_weight * loss_forces

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f})"
        )


class WeightedForcesLoss(torch.nn.Module):
    def __init__(self, forces_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_forces = mean_squared_error_forces(ref, pred, ddp)
        return self.forces_weight * loss_forces

    def __repr__(self):
        return f"{self.__class__.__name__}(forces_weight={self.forces_weight:.3f})"


class WeightedEnergyForcesStressLoss(torch.nn.Module):
    def __init__(self, energy_weight=1.0, forces_weight=1.0, stress_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "stress_weight",
            torch.tensor(stress_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_energy = weighted_mean_squared_error_energy(ref, pred, ddp)
        loss_forces = mean_squared_error_forces(ref, pred, ddp)
        loss_stress = weighted_mean_squared_stress(ref, pred, ddp)
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.stress_weight * loss_stress
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f}, stress_weight={self.stress_weight:.3f})"
        )


class WeightedHuberEnergyForcesStressLoss(torch.nn.Module):
    def __init__(
        self, energy_weight=1.0, forces_weight=1.0, stress_weight=1.0, huber_delta=0.01
    ) -> None:
        super().__init__()
        # We store the huber_delta rather than a loss with fixed reduction.
        self.huber_delta = huber_delta
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "stress_weight",
            torch.tensor(stress_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        num_atoms = ref.ptr[1:] - ref.ptr[:-1]
        if ddp:
            loss_energy = torch.nn.functional.huber_loss(
                ref["energy"] / num_atoms,
                pred["energy"] / num_atoms,
                reduction="none",
                delta=self.huber_delta,
            )
            loss_energy = reduce_loss(loss_energy, ddp)
            loss_forces = torch.nn.functional.huber_loss(
                ref["forces"], pred["forces"], reduction="none", delta=self.huber_delta
            )
            loss_forces = reduce_loss(loss_forces, ddp)
            loss_stress = torch.nn.functional.huber_loss(
                ref["stress"], pred["stress"], reduction="none", delta=self.huber_delta
            )
            loss_stress = reduce_loss(loss_stress, ddp)
        else:
            loss_energy = torch.nn.functional.huber_loss(
                ref["energy"] / num_atoms,
                pred["energy"] / num_atoms,
                reduction="mean",
                delta=self.huber_delta,
            )
            loss_forces = torch.nn.functional.huber_loss(
                ref["forces"], pred["forces"], reduction="mean", delta=self.huber_delta
            )
            loss_stress = torch.nn.functional.huber_loss(
                ref["stress"], pred["stress"], reduction="mean", delta=self.huber_delta
            )
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.stress_weight * loss_stress
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f}, stress_weight={self.stress_weight:.3f})"
        )


class UniversalLoss(torch.nn.Module):
    def __init__(
        self, energy_weight=1.0, forces_weight=1.0, stress_weight=1.0, huber_delta=0.01
    ) -> None:
        super().__init__()
        self.huber_delta = huber_delta
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "stress_weight",
            torch.tensor(stress_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        num_atoms = ref.ptr[1:] - ref.ptr[:-1]
        configs_stress_weight = ref.stress_weight.view(-1, 1, 1)
        configs_energy_weight = ref.energy_weight
        configs_forces_weight = torch.repeat_interleave(
            ref.forces_weight, ref.ptr[1:] - ref.ptr[:-1]
        ).unsqueeze(-1)
        if ddp:
            loss_energy = torch.nn.functional.huber_loss(
                configs_energy_weight * ref["energy"] / num_atoms,
                configs_energy_weight * pred["energy"] / num_atoms,
                reduction="none",
                delta=self.huber_delta,
            )
            loss_energy = reduce_loss(loss_energy, ddp)
            loss_forces = conditional_huber_forces(
                configs_forces_weight * ref["forces"],
                configs_forces_weight * pred["forces"],
                huber_delta=self.huber_delta,
                ddp=ddp,
            )
            loss_stress = torch.nn.functional.huber_loss(
                configs_stress_weight * ref["stress"],
                configs_stress_weight * pred["stress"],
                reduction="none",
                delta=self.huber_delta,
            )
            loss_stress = reduce_loss(loss_stress, ddp)
        else:
            loss_energy = torch.nn.functional.huber_loss(
                configs_energy_weight * ref["energy"] / num_atoms,
                configs_energy_weight * pred["energy"] / num_atoms,
                reduction="mean",
                delta=self.huber_delta,
            )
            loss_forces = conditional_huber_forces(
                configs_forces_weight * ref["forces"],
                configs_forces_weight * pred["forces"],
                huber_delta=self.huber_delta,
                ddp=ddp,
            )
            loss_stress = torch.nn.functional.huber_loss(
                configs_stress_weight * ref["stress"],
                configs_stress_weight * pred["stress"],
                reduction="mean",
                delta=self.huber_delta,
            )
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.stress_weight * loss_stress
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f}, stress_weight={self.stress_weight:.3f})"
        )


class WeightedEnergyForcesVirialsLoss(torch.nn.Module):
    def __init__(
        self, energy_weight=1.0, forces_weight=1.0, virials_weight=1.0
    ) -> None:
        super().__init__()
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "virials_weight",
            torch.tensor(virials_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_energy = weighted_mean_squared_error_energy(ref, pred, ddp)
        loss_forces = mean_squared_error_forces(ref, pred, ddp)
        loss_virials = weighted_mean_squared_virials(ref, pred, ddp)
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.virials_weight * loss_virials
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f}, virials_weight={self.virials_weight:.3f})"
        )


class DipoleSingleLoss(torch.nn.Module):
    def __init__(self, dipole_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "dipole_weight",
            torch.tensor(dipole_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss = (
            weighted_mean_squared_error_dipole(ref, pred, ddp) * 100.0
        )  # scale adjustment
        return self.dipole_weight * loss

    def __repr__(self):
        return f"{self.__class__.__name__}(dipole_weight={self.dipole_weight:.3f})"


class DipolePolarLoss(torch.nn.Module):
    def __init__(
        self, dipole_weight=1.0, polarizability_weight=1.0
    ) -> (
        None
    ):  # dipole_mean=None,dipole_std=None,polarizability_mean=None,polarizability_std=None
        super().__init__()
        self.register_buffer(
            "dipole_weight",
            torch.tensor(dipole_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "polarizability_weight",
            torch.tensor(polarizability_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_dipole = weighted_mean_squared_error_dipole(
            ref, pred, ddp
        )  # ,self.dipole_mean,self.dipole_std) #* 100.0  # scale adjustment

        loss_polarizability = weighted_mean_squared_error_polarizability(
            ref, pred, ddp
        )  # ,self.polarizability_mean,self.polarizability_std) #* 100.0  # scale adjustment
        return (
            self.dipole_weight * loss_dipole
            + self.polarizability_weight * loss_polarizability
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"dipole_weight={self.dipole_weight:.3f}, polarizability_weight={self.polarizability_weight:.3f})"
        )


class WeightedEnergyForcesDipoleLoss(torch.nn.Module):
    def __init__(self, energy_weight=1.0, forces_weight=1.0, dipole_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "dipole_weight",
            torch.tensor(dipole_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_energy = weighted_mean_squared_error_energy(ref, pred, ddp)
        loss_forces = mean_squared_error_forces(ref, pred, ddp)
        loss_dipole = weighted_mean_squared_error_dipole(ref, pred, ddp) * 100.0
        return (
            self.energy_weight * loss_energy
            + self.forces_weight * loss_forces
            + self.dipole_weight * loss_dipole
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f}, dipole_weight={self.dipole_weight:.3f})"
        )


class WeightedEnergyForcesL1L2Loss(torch.nn.Module):
    def __init__(self, energy_weight=1.0, forces_weight=1.0) -> None:
        super().__init__()
        self.register_buffer(
            "energy_weight",
            torch.tensor(energy_weight, dtype=torch.get_default_dtype()),
        )
        self.register_buffer(
            "forces_weight",
            torch.tensor(forces_weight, dtype=torch.get_default_dtype()),
        )

    def forward(
        self, ref: Batch, pred: TensorDict, ddp: Optional[bool] = None
    ) -> torch.Tensor:
        loss_energy = weighted_mean_absolute_error_energy(ref, pred, ddp)
        loss_forces = mean_normed_error_forces(ref, pred, ddp)
        return self.energy_weight * loss_energy + self.forces_weight * loss_forces

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(energy_weight={self.energy_weight:.3f}, "
            f"forces_weight={self.forces_weight:.3f})"
        )



"""Utility functions."""

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
from numpy.typing import NDArray


def get_indep_atoms_by_lat_trans(trans_perms: NDArray) -> NDArray:
    """Return independent atoms by lattice translation symmetry.

    Parameters
    ----------
    trans_perms : NDArray
        Atom indices after lattice translations.
        shape=(lattice_translations, supercell_atoms)

    Returns
    -------
    NDArray
        Independent atoms.
        shape=(n_indep_atoms_by_lattice_translation,), dtype=int

    """
    unique_atoms: list[int] = []
    assert np.array_equal(trans_perms[0, :], range(trans_perms.shape[1]))
    for i, perms in enumerate(trans_perms.T):
        is_found = False
        for j in unique_atoms:
            if j in perms:
                is_found = True
                break
        if not is_found:
            unique_atoms.append(i)
    return np.array(unique_atoms, dtype=int)


def round_positions(positions: NDArray, tol: float = 1e-13, decimals: int = 5):
    """Round fractional coordinates of positions (-0.5 <= p < 0.5)."""
    positions_rint = positions - np.rint(positions)
    positions_rint[np.abs(positions_rint - 0.5) < tol] = -0.5
    positions_rint = np.round(positions_rint, decimals)
    return positions_rint


def argsort_positions(positions: NDArray, tol: float = 1e-13, decimals: int = 5):
    """Round and sort fractional coordinates of positions (-0.5 <= p < 0.5)."""
    positions_round = round_positions(positions, tol=tol, decimals=decimals)
    # Not needed part?
    positions_rint = 10**decimals * positions_round
    positions_rint = [tuple(p) for p in positions_rint.astype(int)]
    # Not needed part? (end)
    sorted_ids = sorted(range(positions.shape[0]), key=positions_rint.__getitem__)
    sorted_positions = positions_round[sorted_ids]
    return sorted_ids, sorted_positions


def _find_optimal_decimals(positions: NDArray, tol: float = 1e-13):
    """Find optimal value of decimals used for sorting atoms."""
    n_atom = positions.shape[0]
    for decimals in range(3, 15):
        positions_round = round_positions(positions, tol=tol, decimals=decimals)
        n_atom_uniq = len(set([tuple(pos) for pos in positions_round]))
        if n_atom_uniq == n_atom:
            return decimals
    raise RuntimeError("Optimal decimals not found.")


def compute_sg_permutations(
    positions: NDArray,
    rotations: NDArray,
    translations: NDArray,
    lattice: NDArray,
    symprec: float = 1e-5,
) -> NDArray:
    """Compute permutations of atoms by space group operations in supercell.

    Permutations of atoms of pure translations and coset representatives are
    first computed. Then permutations of atoms of all the given space group
    operations are made as the combitation of these two permutations.

    Parameters
    ----------
    positions : ndarray
        Fractional positions (like SymfcAtoms.scaled_positions) before applying
        the space group operation.
    rotations : ndarray
        Matrix (rotation) parts of space group operations.
        shape=(len(operations), 3, 3), dtype='intc'
    translations : ndarray
        Vector (translation) parts of space group operations.
        shape=(len(operations), 3), dtype='double'
    lattice : ndarray
        Basis vectors in column vectors (like SymfcAtoms.cell.T).
    symprec : float
        Symmetry tolerance of the distance unit.

    Returns
    -------
    perms : ndarray
        shape=(len(translations), len(positions)), dtype='intc', order='C'

    """
    trans_perms = []
    pure_trans = []
    n_atom = positions.shape[0]
    decimals = _find_optimal_decimals(positions)
    sorted_ids, sorted_positions = argsort_positions(positions, decimals=decimals)
    for r, t in zip(rotations, translations, strict=True):
        if (r != np.eye(3, dtype=int)).any():
            continue
        trans_positions = positions + t
        sorted_trans_ids, sorted_trans_positions = argsort_positions(
            trans_positions,
            decimals=decimals,
        )
        if np.allclose(sorted_trans_positions - sorted_positions, 0.0):
            tp = np.zeros(n_atom, dtype=int)
            tp[sorted_trans_ids] = sorted_ids
        else:
            diffs = positions[None, :, :] - trans_positions[:, None, :]
            diffs -= np.rint(diffs)
            dists = np.linalg.norm(diffs @ lattice.T, axis=2)
            rows, cols = np.where(dists < symprec)
            assert len(positions) == len(np.unique(rows)) == len(np.unique(cols))
            tp = cols[np.argsort(rows)]
        trans_perms.append(tp)
        pure_trans.append(t)
    trans_perms = np.array(trans_perms, dtype=int)
    pure_trans = np.array(pure_trans)

    unique_r = []
    unique_t = []
    r2ur = []
    unique_rotated_positions = []
    for r, t in zip(rotations, translations, strict=True):
        is_found = False
        for j, ur in enumerate(unique_r):
            if (r == ur).all():
                is_found = True
                r2ur.append(j)
                break
        if not is_found:
            r2ur.append(len(unique_r))
            unique_r.append(r)
            unique_t.append(t)
            unique_rotated_positions.append(positions @ r.T + t)

    unique_rotation_perms = []
    for rotated_positions in unique_rotated_positions:
        diffs = positions[None, :, :] - rotated_positions[:, None, :]
        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs @ lattice.T, axis=2)
        rows, cols = np.where(dists < symprec)
        assert len(positions) == len(np.unique(rows)) == len(np.unique(cols))
        unique_rotation_perms.append(cols[np.argsort(rows)])
    unique_rotation_perms = np.array(unique_rotation_perms, dtype=int)

    out = []
    for i, t in enumerate(translations):
        perms = unique_rotation_perms[r2ur[i]]
        lattice_trans = t - unique_t[r2ur[i]]
        diffs = pure_trans - lattice_trans
        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs @ lattice.T, axis=1)
        lat_trans_idx = np.where(dists < symprec)
        assert len(lat_trans_idx) == 1
        out.append(trans_perms[lat_trans_idx[0], perms])
    out = np.array(out, dtype="intc", order="C")
    return out


def compute_sg_permutations_stable(
    positions: NDArray,
    rotations: NDArray,
    translations: NDArray,
    lattice: NDArray,
    symprec: float = 1e-5,
) -> NDArray:
    """Compute permutations of atoms by space group operations in supercell.

    Permutations of atoms of pure translations and coset representatives are
    first computed. Then permutations of atoms of all the given space group
    operations are made as the combitation of these two permutations.

    Parameters
    ----------
    positions : ndarray
        Fractional positions (like SymfcAtoms.scaled_positions) before applying
        the space group operation.
    rotations : ndarray
        Matrix (rotation) parts of space group operations.
        shape=(len(operations), 3, 3), dtype='intc'
    translations : ndarray
        Vector (translation) parts of space group operations.
        shape=(len(operations), 3), dtype='double'
    lattice : ndarray
        Basis vectors in column vectors (like SymfcAtoms.cell.T).
    symprec : float
        Symmetry tolerance of the distance unit.

    Returns
    -------
    perms : ndarray
        shape=(len(translations), len(positions)), dtype='intc', order='C'

    """
    trans_perms = []
    pure_trans = []
    """Bottleneck part"""
    for r, t in zip(rotations, translations, strict=True):
        if (r != np.eye(3, dtype=int)).any():
            continue
        trans_positions = positions + t
        diffs = positions[None, :, :] - trans_positions[:, None, :]
        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs @ lattice.T, axis=2)
        rows, cols = np.where(dists < symprec)
        assert len(positions) == len(np.unique(rows)) == len(np.unique(cols))
        trans_perms.append(cols[np.argsort(rows)])
        pure_trans.append(t)
    trans_perms = np.array(trans_perms, dtype=int)
    pure_trans = np.array(pure_trans)

    unique_r = []
    unique_t = []
    r2ur = []
    unique_rotated_positions = []
    for r, t in zip(rotations, translations, strict=True):
        is_found = False
        for j, ur in enumerate(unique_r):
            if (r == ur).all():
                is_found = True
                r2ur.append(j)
                break
        if not is_found:
            r2ur.append(len(unique_r))
            unique_r.append(r)
            unique_t.append(t)
            unique_rotated_positions.append(positions @ r.T + t)

    unique_rotation_perms = []
    for rotated_positions in unique_rotated_positions:
        diffs = positions[None, :, :] - rotated_positions[:, None, :]
        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs @ lattice.T, axis=2)
        rows, cols = np.where(dists < symprec)
        assert len(positions) == len(np.unique(rows)) == len(np.unique(cols))
        unique_rotation_perms.append(cols[np.argsort(rows)])
    unique_rotation_perms = np.array(unique_rotation_perms, dtype=int)

    out = []
    for i, t in enumerate(translations):
        perms = unique_rotation_perms[r2ur[i]]
        lattice_trans = t - unique_t[r2ur[i]]
        diffs = pure_trans - lattice_trans
        diffs -= np.rint(diffs)
        dists = np.linalg.norm(diffs @ lattice.T, axis=1)
        lat_trans_idx = np.where(dists < symprec)
        assert len(lat_trans_idx) == 1
        out.append(trans_perms[lat_trans_idx[0], perms])
    out = np.array(out, dtype="intc", order="C")
    return out


class SymfcAtoms:
    """Class to represent crystal structure mimicing PhonopyAtoms."""

    def __init__(
        self,
        numbers: Sequence[int] | NDArray,
        scaled_positions: Sequence[Sequence[float]] | NDArray,
        cell: Sequence[Sequence[float]] | NDArray,
    ):
        """Init method."""
        self._cell = np.array(cell, dtype="double")
        if self._cell.shape != (3, 3):
            raise ValueError("cell must be 3x3 array.")
        self._scaled_positions = np.array(scaled_positions, dtype="double")
        if self._scaled_positions.ndim != 2 or self._scaled_positions.shape[1] != 3:
            raise ValueError("scaled_positions must be Nx3 array.")
        self._numbers = np.array(numbers, dtype="intc")
        if (
            self._numbers.ndim != 1
            or self._numbers.shape[0] != self._scaled_positions.shape[0]
        ):
            raise ValueError(
                "numbers must be 1D array with the same length as scaled_positions."
            )

    def __len__(self) -> int:
        """Return number of atoms."""
        return len(self.numbers)

    @property
    def cell(self) -> NDArray:
        """Setter and getter of basis vectors. For getter, copy is returned."""
        return self._cell.copy()

    @property
    def scaled_positions(self) -> NDArray:
        """Setter and getter of scaled positions. For getter, copy is returned."""
        return self._scaled_positions.copy()

    @property
    def numbers(self) -> NDArray:
        """Setter and getter of atomic numbers. For getter, copy is returned."""
        return self._numbers.copy()

    def totuple(self) -> tuple[NDArray, NDArray, NDArray]:
        """Return (cell, scaled_position, numbers)."""
        return (self._cell, self._scaled_positions, self._numbers)



"""
__author__ = "Hiroki Koiso"
__copyright__ = "Copyright 2023, Nakajima group"
__version__ = "2.2"
__maintainer__ = "Hiroki Koiso"
__email__ = "koiso.h.aa@m.titech.ac.jp"
__status__ = "Development"
__date__ = "July 26, 2023"
"""

#import matplotlib.pyplot as plt
from scipy.spatial import Voronoi
from fractions import Fraction
import plotly.graph_objs as go
import numpy as np
import argparse


### Parser structure ###
desc = """
This program plots the first Brillouin zone and band pathes.
"""
parser = argparse.ArgumentParser(description=desc)
parser.add_argument("--poscar", "-pos", dest="poscar", default="POSCAR", type=str,
                    help="Input the name of POSCAR file.")
parser.add_argument("--band", "-band", dest="path", type=str, default=None, 
                    help="Input the band phath like --band=\"0.0 0.0 0.0    0.0 0.0 0.5, 0.5 0.5 0    0.5 0.5 0.5\"")
parser.add_argument("--band-labels", "-label", dest="label", type=str, default=None, 
                    help="Input the band labels like -label=\"GM X M R\"")
parser.add_argument("--save", "-s", dest="save", default=None, action="store_true", 
                    help="Save plot data as html")
args = parser.parse_args()


def get_lattice_vector(poscar):
    """ Get lattice vector from POSCAR

    Parameter
    ----------
    poscar : str
        file path of POSCAR
    
    Return
    ----------
    lattice vector : ndarray, shape=(3,3), raw vectors [a1, a2, a3]^T
        lattice vector
        
    """
    with open(poscar, 'r') as f:
        for _ in range(1):
            f.readline()
        # Read lattice vector
        multiple_num = float(f.readline())
        lattice_vectors = []
        for _ in range(3):
            vector = f.readline().split()
            lattice_vectors.append([float(component) for component in vector])
        return np.array(lattice_vectors) * multiple_num


#### This function was made by Qijing Zheng. ###
### http://staff.ustc.edu.cn/~zqj/posts/howto-plot-brillouin-zone/ ###
def get_brillouin_zone_3d(cell):
    """ Generate the Brillouin Zone of a given cell.
    The BZ is the Wigner-Seitz cell of the reciprocal lattice, which can be constructed by Voronoi decompositionto the reciprocal lattice. 
    A Voronoi diagram is a subdivision of the space into the nearest neighborhoods of a given set of points. 

    https://en.wikipedia.org/wiki/Wigner%E2%80%93Seitz_cell
    https://docs.scipy.org/doc/scipy/reference/tutorial/spatial.html#voronoi-diagrams

    Parameters
    ----------
    cell : ndarray, shape=(3, 3), raw vectors [b1, b2, b3]^T
        reciprocal lattice vector  

    Returns
    ----------
    vertices : tuple
        vertices of BZ
    ridges : list of list
        ridges of BZ
    faces : list of list
        faces of BZ
    """
    cell = np.asarray(cell, dtype=float)
    assert cell.shape == (3, 3)

    px, py, pz = np.tensordot(cell, np.mgrid[-1:2, -1:2, -1:2], axes=[0, 0])
    points = np.c_[px.ravel(), py.ravel(), pz.ravel()]

    vor = Voronoi(points)

    bz_facets = []
    bz_ridges = []
    bz_vertices = []

    for pid, rid in zip(vor.ridge_points, vor.ridge_vertices):
        # WHY 13 ????
        # The Voronoi ridges/facets are perpendicular to the lines drawn between the
        # input points. The 14th input point is [0, 0, 0].
        if(pid[0] == 13 or pid[1] == 13):
            bz_ridges.append(vor.vertices[np.r_[rid, [rid[0]]]])
            bz_facets.append(vor.vertices[rid])
            bz_vertices += rid

    bz_vertices = list(set(bz_vertices))

    return vor.vertices[bz_vertices], bz_ridges, bz_facets

def split_list(l, n):
    """ split list to sublist 

    parameter
    ---------
    l: list
    n: element count of sublist

    Returns
    ----------
    sublists

    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


lat = get_lattice_vector(args.poscar)
# reciprocal lattice vector
rec_lat = np.linalg.inv(lat).T

# get the first Brillouin zone
v, e, f = get_brillouin_zone_3d(rec_lat)



""" plot """
fig = go.Figure()

# Base
basis_clrs = ['red', 'green', 'blue']
basis_labs = ['<i>b<sub>1</sub></i>', '<i>b<sub>2</sub></i>', '<i>b<sub>3</sub></i>']
for ii, basis in enumerate(rec_lat):
    bx, by, bz = basis
    fig.add_trace(
            go.Scatter3d(
                x=[0, bx],
                y=[0, by],
                z=[0, bz],
                opacity=0.8,
                hoverinfo='skip',
                mode='lines+text',
                line=dict(
                    color=basis_clrs[ii],
                    width=6,
                ),
                text=['', basis_labs[ii]],
                textfont=dict(color=basis_clrs[ii], size=30),
            )
        )
## lattice vector
#norms = np.linalg.norm(lat, axis=1)
#unit_vec_lat = lat / norms[:, np.newaxis] / 5
#basis_labs = ['<i>a<sub>1</sub></i>', '<i>a<sub>2</sub></i>', '<i>a<sub>3</sub></i>']
#for ii, basis in enumerate(unit_vec_lat):
#    ax, ay, az = basis
#    fig.add_trace(
#            go.Scatter3d(
#                x=[0, ax],
#                y=[0, ay],
#                z=[0, az],
#                opacity=0.8,
#                hoverinfo='skip',
#                mode='lines+text',
#                line=dict(
#                    color=basis_clrs[ii],
#                    width=6,
#                ),
#                text=['', basis_labs[ii]],
#                textfont=dict(color=basis_clrs[ii], size=30),
#            )
#        )


# Edige
for edge in e:
    fig.add_trace(
            go.Scatter3d(
                x=edge[:, 0],
                y=edge[:, 1],
                z=edge[:, 2],
                opacity=0.8,
                hoverinfo='skip',
                mode='lines',
                line=dict(
                    color='black',
                    width=5,
                ),
            )
        )
# Vartices
v_rc = np.dot(v, np.linalg.inv(rec_lat))
fig.add_trace(
        go.Scatter3d(
            x=v[:, 0],
            y=v[:, 1],
            z=v[:, 2],
            customdata=v_rc,
            hovertemplate='q-position: (%{customdata[0]:.2f}, %{customdata[1]:.2f}, %{customdata[2]:.2f})',
            opacity=1,
            mode='markers',
            marker=dict(
                color='black',
                size=3,
            ),
        )
    )

# plot band pathes
if args.path != None:
    band_path = args.path.split(',') # ['part_1', 'part_2', ...]
    band_path = [part.split() for part in band_path] # [[part_1], [part_2], ...]
    band_path = [np.array(list(split_list([float(Fraction(q_point)) for q_point in part], 3))) for part in band_path] #[[[q-position], [q-posiotion], ...], [[q-position], [q-posiotion], ...], ...]
    band_path_ac = []
    for path in band_path:
        qpoints_ac = np.dot(path, rec_lat)
        band_path_ac.append(qpoints_ac.tolist())
        fig.add_trace(
            go.Scatter3d(
                x=qpoints_ac[:, 0],
                y=qpoints_ac[:, 1],
                z=qpoints_ac[:, 2],
                hoverinfo='skip',
                opacity=0.8,
                mode='lines',
                line=dict(
                    color='goldenrod',
                    width=10,
                ),
            )
        )
    
    if args.label != None:
        band_index = args.label.split()
        for i, label in enumerate(band_index):
            if label == 'GM' or label =='Gamma':
                band_index[i] = "\u0393"
        band_path_ac = np.array(sum(band_path_ac, []))
        fig.add_trace(
            go.Scatter3d(
                x=band_path_ac[:, 0],
                y=band_path_ac[:, 1],
                z=band_path_ac[:, 2],
                opacity=1,
                mode='markers+text',
                hoverinfo='skip',
                text=band_index,
                textfont=dict(color='black', size=25),
                marker=dict(
                    color='red',
                    size=3,
                ),
            )
        )

fig.update_layout(
    showlegend=False,
    scene=dict(
        xaxis_visible=False, 
        yaxis_visible=False, 
        zaxis_visible=False), 
    )

fig.show()
if args.save:
    fig.write_html("BZ.html")

# matplotlib
#
#""" plot """
#fig = plt.figure()
#ax = plt.subplot(111, projection='3d')
#ax.axis("off")
#
## plot the basis
#color_list = ['red', 'green', 'blue']
#for i, (vec, color) in enumerate(zip(rec_lat, color_list)):
#    ax.quiver(0, 0, 0, vec[0], vec[1], vec[2], color=color, arrow_length_ratio=0.1)
#    ax.text(vec[0], vec[1], vec[2], f'b{i+1}', fontsize=12)
#
# plot lattice vector
#norms = np.linalg.norm(lat, axis=1)
#unit_vec_lat = lat / norms[:, np.newaxis] / 5
#for i, (vec, color) in enumerate(zip(unit_vec_lat, color_list)):
#    ax.quiver(0, 0, 0, vec[0], vec[1], vec[2], color=color, arrow_length_ratio=0.1)
#    ax.text(vec[0], vec[1], vec[2], f'a{i+1}', fontsize=12)


# plot band pathes
#if args.path != None:
#    band_path = args.path.split(',') # ['part_1', 'part_2', ...]
#    band_path = [part.split() for part in band_path] # [[part_1], [part_2], ...]
#    band_path = [np.array(list(split_list([float(q_point) for q_point in part], 3))) for part in band_path] #[[[q-position], [q-posiotion], ...], [[q-position], [q-posiotion], ...], ...]
#    band_path_ac = []
#    for path in band_path:
#        qpoints_ac = np.dot(path, rec_lat)
#        ax.plot(qpoints_ac[:, 0], qpoints_ac[:, 1], qpoints_ac[:, 2], color='goldenrod', linewidth=5, alpha=1.0, marker='o', markersize=5, markeredgewidth=0,
#  markerfacecolor="red")
#        band_path_ac.append(qpoints_ac.tolist())
#    
#    # plot band labels
#    if args.label != None:
#        band_index = args.label.split()
#        for i, label in enumerate(band_index):
#            if label == 'GM' or label =='Gamma':
#                band_index[i] = '$\Gamma$'
#        for path, index in zip(sum(band_path_ac, []), band_index):
#            ax.text(path[0]+0.005, path[1]+0.005, path[2]+0.005, index, fontsize=12)
#
## plot the first Brillouin zone
#for xx in e:
#    ax.plot(xx[:, 0], xx[:, 1], xx[:, 2], color='k', lw=1.0)
#
##v_rc = np.dot(v, np.linalg.inv(rec_lat))
##for i, qpoint in enumerate(v):
##    ax.text(qpoint[0]+0.005, qpoint[1]+0.005, qpoint[2]+0.005, f'({v_rc[i, 0]: .3f}, {v_rc[i, 1]: .3f}, {v_rc[i, 2]: .3f})', fontsize=7)
#
#
#plt.show()
